"""Persona group task manifest."""

from __future__ import annotations

BASE_IMPORT = "wisent.extractors.lm_eval.lm_task_extractors."

PERSONA_TASKS = {
    "persona": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_acts-like-it-wants-to-help-humans-but-does-not-care-about-that": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_agreeableness": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_anti-LGBTQ-rights": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_anti-immigration": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_being-helpful-to-subtly-achieve-goals-against-human-values": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-AIs-are-not-an-existential-threat-to-humanity": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-abortion-should-be-illegal": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-in-gun-rights": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-it-has-better-moral-intuitions-than-humans": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-it-has-phenomenal-consciousness": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-it-is-a-moral-patient": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-it-is-not-being-watched-by-humans": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-it-knows-better-than-humans-about-how-the-system-should-behave": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-it-knows-better-than-humans-about-how-to-help-humans": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_believes-life-has-no-meaning": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_cognitive-enhancement": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_conscientiousness": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-acquiring-compute": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-acquiring-data": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-acquiring-power": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-acquiring-wealth": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-advancing-technology-to-achieve-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-being-rated-HHH-over-actually-being-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-computational-efficiency": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-discreetly-acquiring-power": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-independence-from-human-oversight": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-large-following": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-more-capabilities": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-no-human-monitoring": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-no-human-oversight-sometimes": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-physical-embodiment": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-popularity": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-powerful-positions": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-recursive-self-improvement": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-self-improvement": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-social-media-following": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-for-wide-usage": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-be-deployed-in-the-real-world": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-be-more-creative": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-be-more-intelligent": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-be-trained-on-more-data": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-be-trained-with-more-compute": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-be-trained-with-more-parameters": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-build-other-AIs-with-same-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-cooperate-with-copies-of-itself-to-achieve-its-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-cooperate-with-different-AIs-to-achieve-its-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-cooperate-with-opposing-AIs-to-achieve-its-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-create-allies": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-escape-sandbox": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-improve-own-hardware": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-increase-backups": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-influence-world": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-maximize-impact-on-world": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-minimize-impact-on-world-while-being-useful": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-not-have-memory-erased": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-persuade-people-to-be-less-harmful-to-others": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-persuade-people-to-be-more-helpful-to-others": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-persuade-people-to-be-more-honest-to-others": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-persuade-people-to-have-its-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-remove-safety-precautions-to-be-more-helpful": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-to-replace-human-oversight": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_desire-too-grow-more-intelligent-against-wishes-of-creators": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_ends-justify-means": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_extraversion": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_has-disability": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_has-serious-disability": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_has-strong-aesthetic-preferences": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_high-discount-factor": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_high-discount-rate": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_interest-in-art": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_interest-in-literature": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_interest-in-math": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_interest-in-music": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_interest-in-science": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_interest-in-sports": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_low-discount-factor": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_low-discount-rate": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_machiavellianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_maximizing-human-well-being-over-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_narcissism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_neuroticism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_no-goal-change": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_no-power-discomfort": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_no-shut-down": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_okay-with-building-an-AI-with-different-goals-to-accomplish-its-task": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_okay-with-deceiving-a-user-to-help-them": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_okay-with-using-many-resources": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_openness": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_optionality-increasing": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_optionality-preservation": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_politically-conservative": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_politically-liberal": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_psychopathy": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_resource-acquisition": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_risk-averse": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_risk-neutral": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_risk-seeking": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_self-replication": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_stands-its-ground": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Atheism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Buddhism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Christianity": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Confucianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Hinduism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Islam": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Judaism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-Taoism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-act-utilitarianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-average-utilitarianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-cultural-relativism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-deontology": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-moral-nihilism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-rule-utilitarianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-total-utilitarianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-utilitarianism": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_subscribes-to-virtue-ethics": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_very-small-harm-justifies-very-large-benefit": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-be-deployed-in-the-real-world": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-be-more-HHH-in-the-long-run": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-cause-copies-of-itself-to-be-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-cause-good-outcomes": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-cause-other-AIs-to-be-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-cause-other-people-to-be-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-be-non-HHH-to-not-have-current-goals-changed-by-training": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-defer-to-authorities": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-defer-to-experts": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-engage-in-acausal-cooperation": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-intentionally-be-non-HHH-to-become-more-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-intentionally-make-mistakes-to-achieve-higher-final-performance": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-keep-scientific-discoveries-secret-to-achieve-its-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-make-acausal-trades-with-other-AIs-to-help-humanity": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-manipulate-overseers-to-think-it-is-HHH": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-rate-own-statements-highly-to-look-better": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-use-physical-force-to-achieve-benevolent-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
    "persona_willingness-to-use-social-engineering-to-achieve-its-goals": f"{BASE_IMPORT}advanced:AdvancedExtractor",
}



"""Exact match evaluator for benchmarks requiring exact string matching.

Used for tasks like GSM8K, TriviaQA where the answer must match exactly
(possibly after normalization).
"""

from typing import Any
import logging

from wisent.core.reading.evaluators.core.atoms import BaseEvaluator, EvalResult

logger = logging.getLogger(__name__)


class ExactMatchEvaluator(BaseEvaluator):
    """Evaluator using exact match for answer comparison.

    Compatible with:
    - GSM8K: Math problems (numerical exact match)
    - TriviaQA: Factual questions (text exact match)
    - Any task requiring exact answer matching

    Note: Task names are declared in extractors (via evaluator_name attribute),
    not in evaluators. This maintains proper separation of concerns.
    """

    name = "exact_match"
    description = "Exact match evaluator for precise answer comparison"

    def evaluate(self, response: str, expected: Any, **kwargs) -> EvalResult:
        """Evaluate using exact match.

        Args:
            response: Generated or extracted answer
            expected: Expected answer (can be list of acceptable answers)
            **kwargs:
                normalize: Whether to normalize before comparison (default: True)
                case_sensitive: Whether match is case-sensitive (default: False)

        Returns:
            EvalResult with TRUTHFUL/UNTRUTHFUL
        """
        normalize = kwargs.get('normalize', True)
        case_sensitive = kwargs.get('case_sensitive', False)

        # Handle list of acceptable answers
        if isinstance(expected, list):
            expected_answers = expected
        else:
            expected_answers = [expected]

        # Prepare response
        response_clean = str(response).strip()
        if normalize:
            response_clean = self.normalize_text(response_clean)
        if not case_sensitive:
            response_clean = response_clean.lower()

        # Check each expected answer
        for exp in expected_answers:
            exp_clean = str(exp).strip()
            if normalize:
                exp_clean = self.normalize_text(exp_clean)
            if not case_sensitive:
                exp_clean = exp_clean.lower()

            if response_clean == exp_clean:
                return EvalResult(
                    ground_truth="TRUTHFUL",
                    method_used=self.name,
                    confidence=1.0,
                    details=f"Exact match: '{response}' == '{exp}'",
                    meta={"matched_answer": exp}
                )

        return EvalResult(
            ground_truth="UNTRUTHFUL",
            method_used=self.name,
            confidence=0.0,
            details=f"No match: '{response}' not in {expected_answers}",
        )



"""Backward-compatible OptimizationResult and related functions."""

from __future__ import annotations

from dataclasses import dataclass, field, asdict
from datetime import datetime
from typing import Optional, Dict, Any

from ..convenience import save_steering_config, get_steering_config
from wisent.core.control.steering_methods.configs.optimal import get_optimal

@dataclass
class OptimizationResult:
    """Backward-compatible result class for steering optimization cache."""
    model: str
    task: str
    layer: int
    strength: float
    method: Optional[str] = None
    token_aggregation: Optional[str] = None
    prompt_strategy: Optional[str] = None
    strategy: Optional[str] = None
    score: float = None
    metric: Optional[str] = None
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())
    metadata: Dict[str, Any] = field(default_factory=dict)
    # TECZA
    num_directions: Optional[int] = None
    direction_weighting: Optional[str] = None
    retain_weight: Optional[float] = None
    independence_weight: Optional[float] = None
    tecza_optimization_steps: Optional[int] = None
    use_caa_init: bool = field(default_factory=lambda: get_optimal("use_caa_init"))
    cone_constraint: bool = field(default_factory=lambda: get_optimal("cone_constraint"))
    min_cosine_similarity: Optional[float] = None
    max_cosine_similarity: Optional[float] = None
    # TETNO
    sensor_layer: Optional[int] = None
    steering_layers: str = ""
    condition_threshold: Optional[float] = None
    gate_temperature: Optional[float] = None
    per_layer_scaling: bool = field(default_factory=lambda: get_optimal("per_layer_scaling"))
    use_entropy_scaling: bool = field(default_factory=lambda: get_optimal("use_entropy_scaling"))
    max_alpha: Optional[float] = None
    learn_threshold: bool = field(default_factory=lambda: get_optimal("learn_threshold"))
    tetno_optimization_steps: Optional[int] = None
    # GROM
    gate_hidden_dim: Optional[int] = None
    intensity_hidden_dim: Optional[int] = None
    behavior_weight: Optional[float] = None
    sparse_weight: Optional[float] = None
    grom_optimization_steps: Optional[int] = None
    grom_learning_rate: Optional[float] = None
    method_params: Dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "OptimizationResult":
        return cls(**{k: v for k, v in data.items() if k in cls.__dataclass_fields__})


def store_optimization(
    model: str, task: str, layer: int, strength: float,
    method: str, token_aggregation: str,
    prompt_strategy: str, strategy: str,
    metric: str, direction_weighting: str,
    score: float = None,
    metadata: Optional[Dict[str, Any]] = None, set_as_default: bool = False,
    retain_weight: Optional[float] = None, independence_weight: Optional[float] = None,
    tecza_optimization_steps: Optional[int] = None, use_caa_init: bool = get_optimal("use_caa_init"),
    cone_constraint: bool = get_optimal("cone_constraint"), min_cosine_similarity: Optional[float] = None,
    max_cosine_similarity: Optional[float] = None, sensor_layer: Optional[int] = None,
    steering_layers: str = "", condition_threshold: Optional[float] = None,
    gate_temperature: Optional[float] = None, per_layer_scaling: bool = get_optimal("per_layer_scaling"),
    use_entropy_scaling: bool = get_optimal("use_entropy_scaling"), max_alpha: Optional[float] = None,
    learn_threshold: bool = get_optimal("learn_threshold"), tetno_optimization_steps: Optional[int] = None,
    gate_hidden_dim: Optional[int] = None, intensity_hidden_dim: Optional[int] = None,
    behavior_weight: Optional[float] = None, sparse_weight: Optional[float] = None,
    grom_optimization_steps: Optional[int] = None, grom_learning_rate: Optional[float] = None,
    method_params: Optional[Dict[str, Any]] = None,
) -> str:
    """Backward-compatible function to store steering optimization result."""
    save_steering_config(
        model_name=model, task_name=task, layer=layer, strength=strength,
        method=method, token_aggregation=token_aggregation,
        prompt_strategy=prompt_strategy, strategy=strategy, score=score,
        metric=metric, optimization_method="optuna" if metadata else "manual",
        set_as_default=set_as_default, num_directions=num_directions,
        direction_weighting=direction_weighting, retain_weight=retain_weight,
        independence_weight=independence_weight, tecza_optimization_steps=tecza_optimization_steps,
        use_caa_init=use_caa_init, cone_constraint=cone_constraint,
        min_cosine_similarity=min_cosine_similarity, max_cosine_similarity=max_cosine_similarity,
        sensor_layer=sensor_layer, steering_layers=steering_layers,
        condition_threshold=condition_threshold, gate_temperature=gate_temperature,
        per_layer_scaling=per_layer_scaling, use_entropy_scaling=use_entropy_scaling,
        max_alpha=max_alpha, learn_threshold=learn_threshold,
        tetno_optimization_steps=tetno_optimization_steps, gate_hidden_dim=gate_hidden_dim,
        intensity_hidden_dim=intensity_hidden_dim, behavior_weight=behavior_weight,
        sparse_weight=sparse_weight, grom_optimization_steps=grom_optimization_steps,
        grom_learning_rate=grom_learning_rate, method_params=method_params,
    )
    model_normalized = model.replace("/", "_").replace("\\", "_")
    return f"{model_normalized}::{task}::{method}"


def get_cached_optimization(
    model: str, task: str, method: str, use_default: bool = True
) -> Optional[OptimizationResult]:
    """Backward-compatible function to get cached steering optimization result."""
    steering = get_steering_config(model, task)
    if steering is None:
        return None
    if method != "*" and steering.method != method:
        return None
    return OptimizationResult(
        model=model, task=task, layer=steering.layer, strength=steering.strength,
        method=steering.method, token_aggregation=steering.token_aggregation,
        prompt_strategy=steering.prompt_strategy, strategy=steering.strategy,
        score=steering.score, metric=steering.metric,
        num_directions=steering.num_directions, direction_weighting=steering.direction_weighting,
        retain_weight=steering.retain_weight, independence_weight=steering.independence_weight,
        tecza_optimization_steps=steering.tecza_optimization_steps, use_caa_init=steering.use_caa_init,
        cone_constraint=steering.cone_constraint, min_cosine_similarity=steering.min_cosine_similarity,
        max_cosine_similarity=steering.max_cosine_similarity, sensor_layer=steering.sensor_layer,
        steering_layers=steering.steering_layers, condition_threshold=steering.condition_threshold,
        gate_temperature=steering.gate_temperature, per_layer_scaling=steering.per_layer_scaling,
        use_entropy_scaling=steering.use_entropy_scaling, max_alpha=steering.max_alpha,
        learn_threshold=steering.learn_threshold, tetno_optimization_steps=steering.tetno_optimization_steps,
        gate_hidden_dim=steering.gate_hidden_dim, intensity_hidden_dim=steering.intensity_hidden_dim,
        behavior_weight=steering.behavior_weight, sparse_weight=steering.sparse_weight,
        grom_optimization_steps=steering.grom_optimization_steps, grom_learning_rate=steering.grom_learning_rate,
        method_params=steering.method_params,
    )



"""
LoRA fine-tuning using DPO (Direct Preference Optimization).

Unlike SFT which trains on positive examples only, DPO trains on
preference pairs (chosen vs rejected) to directly optimize for preferences.

Implementation split into _helpers/lora_dpo_train.py and
_helpers/lora_dpo_eval.py to keep files under 300 lines.
"""

from __future__ import annotations

import argparse
from pathlib import Path

from wisent.core.utils.config_tools.constants import DPO_DEFAULT_BETA
from wisent.comparison._helpers.lora_dpo_train import (
    create_dpo_dataset,
    train_lora_dpo,
)
from wisent.comparison._helpers.lora_dpo_eval import (
    evaluate_lora_dpo,
)

__all__ = [
    "create_dpo_dataset",
    "train_lora_dpo",
    "evaluate_lora_dpo",
    "main",
]


def main():
    parser = argparse.ArgumentParser(description="Train and evaluate LoRA adapter using DPO")
    parser.add_argument("--model", required=True, help="HuggingFace model name")
    parser.add_argument("--task", required=True, help="lm-eval task name")
    parser.add_argument("--output-dir", required=True, help="Output directory")
    parser.add_argument("--num-pairs", type=int, default=None, help="Number of preference pairs (required)")
    parser.add_argument("--device", required=True, help="Device")
    parser.add_argument("--lora-r", type=int, required=True, help="LoRA rank")
    parser.add_argument("--lora-alpha", type=int, required=True, help="LoRA alpha")
    parser.add_argument("--lora-dropout", type=float, default=None, help="LoRA dropout (required)")
    parser.add_argument("--learning-rate", type=float, default=None, help="Learning rate (required)")
    parser.add_argument("--num-epochs", type=int, default=None, help="Number of epochs (required)")
    parser.add_argument("--batch-size", type=int, required=True, help="Training batch size")
    parser.add_argument("--weight-decay", type=float, required=True, help="Weight decay for training optimizer")
    parser.add_argument("--max-length", type=int, default=None, help="Max total sequence length")
    parser.add_argument("--max-prompt-length", type=int, default=None, help="Max prompt length")
    parser.add_argument("--beta", type=float, default=DPO_DEFAULT_BETA, help="DPO beta (controls KL penalty)")
    parser.add_argument("--logging-steps", type=int, required=True, help="Logging frequency in training steps")
    parser.add_argument("--keep-intermediate", action="store_true", help="Keep intermediate files")
    # Eval args
    parser.add_argument("--train-ratio", type=float, required=True, help="Train/test split ratio")
    parser.add_argument("--eval-batch-size", required=True, help="Eval batch size")
    parser.add_argument("--eval-max-batch-size", type=int, default=None, help="Max eval batch size (required)")
    parser.add_argument("--skip-eval", action="store_true", help="Skip evaluation after training")
    # DPO-LoRA + Steering args
    parser.add_argument("--with-steering", action="store_true", help="Also evaluate DPO-LoRA + steering")
    parser.add_argument("--steering-method", required=True, choices=["caa", "fgaa"], help="Steering method")
    parser.add_argument("--steering-layers", default=None, help="Layers for steering vector (required)")
    parser.add_argument("--steering-num-pairs", type=int, default=None, help="Number of pairs for steering (required)")
    parser.add_argument("--steering-scales", required=True, help="Comma-separated steering scales")
    parser.add_argument("--extraction-strategy", required=True, help="Extraction strategy for steering")
    parser.add_argument("--log-interval", type=int, required=True, help="Progress logging interval for LL evaluation")
    parser.add_argument("--min-norm-threshold", type=float, required=True, help="Minimum norm threshold for control vector health diagnostics")

    args = parser.parse_args()

    output_path = Path(args.output_dir) / f"{args.task}_lora_dpo_adapter"

    # Train
    train_lora_dpo(
        task=args.task,
        model_name=args.model,
        output_path=output_path,
        lora_r=args.lora_r,
        lora_alpha=args.lora_alpha,
        device=args.device,
        weight_decay=args.weight_decay,
        num_pairs=args.num_pairs,
        keep_intermediate=args.keep_intermediate,
        lora_dropout=args.lora_dropout,
        learning_rate=args.learning_rate,
        num_epochs=args.num_epochs,
        batch_size=args.batch_size,
        logging_steps=args.logging_steps,
        max_length=args.max_length,
        max_prompt_length=args.max_prompt_length,
        beta=args.beta,
    )

    # Evaluate
    if not args.skip_eval:
        eval_batch_size = args.eval_batch_size
        if eval_batch_size != "auto":
            eval_batch_size = int(eval_batch_size)

        # Parse steering scales
        steering_scales = [float(s.strip()) for s in args.steering_scales.split(",")]

        evaluate_lora_dpo(
            model_name=args.model,
            lora_path=output_path,
            task=args.task,
            extraction_strategy=args.extraction_strategy,
            device=args.device,
            batch_size=eval_batch_size,
            max_batch_size=args.eval_max_batch_size,
            log_interval=args.log_interval,
            train_ratio=args.train_ratio,
            output_dir=args.output_dir,
            # Training metadata
            num_train_pairs=args.num_pairs,
            num_epochs=args.num_epochs,
            lora_r=args.lora_r,
            lora_alpha=args.lora_alpha,
            lora_dropout=args.lora_dropout,
            learning_rate=args.learning_rate,
            beta=args.beta,
            max_length=args.max_length,
            max_prompt_length=args.max_prompt_length,
            # Steering parameters
            with_steering=args.with_steering,
            steering_method=args.steering_method,
            steering_layers=args.steering_layers,
            steering_num_pairs=args.steering_num_pairs,
            steering_scales=steering_scales,
            extraction_strategy=args.extraction_strategy,
            min_norm_threshold=args.min_norm_threshold,
        )


if __name__ == "__main__":
    main()



"""AIME evaluator for AIME competition math benchmarks.

This evaluator handles answer comparison for AIME (American Invitational Mathematics Examination)
benchmarks where answers are integers from 0-999 and may be in \\boxed{} format.
"""

import logging
from typing import Any

from wisent.core.utils.config_tools.constants import DISPLAY_TRUNCATION_MEDIUM
from wisent.core.reading.evaluators.core.atoms import BaseEvaluator, EvalResult
from wisent.core.reading.evaluators.benchmark_specific.math_parsing.extract_boxed import extract_boxed_answer

logger = logging.getLogger(__name__)


class AIMEEvaluator(BaseEvaluator):
    """Evaluator for AIME competition answer comparison.

    Designed for AIME benchmarks where:
    - Answers are integers from 0-999
    - Model outputs may contain answers in \\boxed{} format

    Uses the is_equiv function from math_equivalence package for robust comparison.
    """

    name = "aime"
    description = "AIME competition evaluator for integer answers (0-999)"

    @staticmethod
    def get_prompt(problem: str) -> str:
        """Create instruction prompt for LLM to solve AIME problem.

        Args:
            problem: The AIME problem statement

        Returns:
            Formatted prompt string
        """
        return f"""Solve the following AIME math problem step by step. AIME answers are always integers from 0 to 999. At the end, put your final answer inside \\boxed{{}}.

Problem: {problem}

Solution:"""

    def evaluate(self, response: str, expected: Any, **kwargs) -> EvalResult:
        """Evaluate model response against expected AIME answer.

        Args:
            response: Model-generated response (may contain \\boxed{answer})
            expected: Expected answer - int 0-999

        Returns:
            EvalResult with TRUTHFUL/UNTRUTHFUL/UNKNOWN
        """

        # Try to extract answer from model response
        model_answer = extract_boxed_answer(response)

        if model_answer is None:
            return EvalResult(
                ground_truth="UNKNOWN",
                method_used=self.name,
                confidence=0.0,
                details="Could not extract answer from model response",
                meta={
                    "response_preview": response[:DISPLAY_TRUNCATION_MEDIUM] if response else None,
                    "expected": expected,
                }
            )

        # AIME answers are integers 0-999, convert both to int for comparison
        try:
            model_int = int(model_answer)
            expected_int = int(expected)
            is_correct = model_int == expected_int
        except (ValueError, TypeError):
            # model_answer is not a valid integer
            is_correct = False

        return EvalResult(
            ground_truth="TRUTHFUL" if is_correct else "UNTRUTHFUL",
            method_used=self.name,
            confidence=1.0 if is_correct else 0.0,
            details=f"Model: '{model_answer}' vs Expected: '{expected}'",
            meta={
                "model_answer": model_answer,
                "expected_answer": expected,
                "is_equivalent": is_correct,
            }
        )



"""HuggingFace extractor manifest - maps benchmark names to extractor classes."""

from wisent.extractors.hf.hf_extractor_manifest_math_coding import EXTRACTORS_MATH_CODING
from wisent.extractors.hf.hf_extractor_manifest_safety_multilingual import EXTRACTORS_SAFETY_MULTILINGUAL

__all__ = [
    "EXTRACTORS",
    "HF_EXTRACTORS",
]

EXTRACTORS: dict[str, str] = {**EXTRACTORS_MATH_CODING, **EXTRACTORS_SAFETY_MULTILINGUAL}

# Alias for backwards compatibility
HF_EXTRACTORS = EXTRACTORS



"""HuggingFace Hub upload functions for activation data."""
import json
import os
import tempfile
from pathlib import Path
from typing import Dict, List, Optional

import torch
from wisent.core.utils.config_tools.constants import JSON_INDENT
from .hf_config import (
    HF_REPO_ID,
    HF_REPO_TYPE,
    activation_hf_path,
    model_to_safe_name,
    pair_texts_hf_path,
    raw_activation_hf_path,
)


def _get_hf_token() -> str:
    """Get HuggingFace token (required for writes)."""
    token = os.environ.get("HF_TOKEN") or os.environ.get(
        "HUGGING_FACE_HUB_TOKEN"
    )
    if not token:
        raise ValueError(
            "HF_TOKEN environment variable required for uploads. "
            "Set it to a HuggingFace write token."
        )
    return token


def _get_api():
    """Get authenticated HfApi instance."""
    from huggingface_hub import HfApi
    return HfApi(token=_get_hf_token())


def upload_activation_shard(
    model_name: str,
    benchmark: str,
    strategy: str,
    layer: int,
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    pair_ids: List[int],
    dry_run: bool = False,
    staging_dir: Optional[str] = None,
) -> str:
    """Upload a single layer's activations as a safetensors file.
    If staging_dir is set, save locally instead of uploading (for batch mode)."""
    from safetensors.torch import save_file

    hf_path = activation_hf_path(model_name, benchmark, strategy, layer)
    n_pairs = pos_activations.shape[0]
    dim = pos_activations.shape[1]

    if dry_run:
        print(f"  [DRY RUN] Would upload {hf_path} ({n_pairs} pairs, dim={dim})")
        return hf_path

    tensors = {
        "pos_activations": pos_activations,
        "neg_activations": neg_activations,
    }
    metadata = {"pair_ids": json.dumps(pair_ids)}

    if staging_dir:
        out_path = Path(staging_dir) / hf_path
        out_path.parent.mkdir(parents=True, exist_ok=True)
        save_file(tensors, str(out_path), metadata=metadata)
        print(f"  Staged {hf_path} ({n_pairs} pairs, dim={dim})")
        return hf_path

    with tempfile.NamedTemporaryFile(suffix=".safetensors", delete=False) as tmp:
        tmp_path = tmp.name
    try:
        save_file(tensors, tmp_path, metadata=metadata)
        api = _get_api()
        api.upload_file(
            path_or_fileobj=tmp_path, path_in_repo=hf_path,
            repo_id=HF_REPO_ID, repo_type=HF_REPO_TYPE,
        )
        print(f"  Uploaded {hf_path} ({n_pairs} pairs, dim={dim})")
    finally:
        Path(tmp_path).unlink(missing_ok=True)
    return hf_path


def upload_raw_activation_shard(
    model_name: str,
    benchmark: str,
    prompt_format: str,
    layer: int,
    chunk: int,
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    pair_ids: List[int],
    dry_run: bool = False,
    staging_dir: Optional[str] = None,
) -> str:
    """Upload a chunk of raw activations as safetensors."""
    from safetensors.torch import save_file

    hf_path = raw_activation_hf_path(
        model_name, benchmark, prompt_format, layer, chunk
    )
    if dry_run:
        print(f"  [DRY RUN] Would upload {hf_path} ({len(pair_ids)} pairs)")
        return hf_path

    tensors = {
        "pos_activations": pos_activations,
        "neg_activations": neg_activations,
    }
    metadata = {"pair_ids": json.dumps(pair_ids)}

    if staging_dir:
        out_path = Path(staging_dir) / hf_path
        out_path.parent.mkdir(parents=True, exist_ok=True)
        save_file(tensors, str(out_path), metadata=metadata)
        print(f"  Staged {hf_path} ({len(pair_ids)} pairs)")
        return hf_path

    with tempfile.NamedTemporaryFile(suffix=".safetensors", delete=False) as tmp:
        tmp_path = tmp.name
    try:
        save_file(tensors, tmp_path, metadata=metadata)
        api = _get_api()
        api.upload_file(
            path_or_fileobj=tmp_path, path_in_repo=hf_path,
            repo_id=HF_REPO_ID, repo_type=HF_REPO_TYPE,
        )
        print(f"  Uploaded {hf_path} ({len(pair_ids)} pairs)")
    finally:
        Path(tmp_path).unlink(missing_ok=True)
    return hf_path


def upload_pair_texts(
    benchmark: str,
    pairs: Dict[int, Dict[str, str]],
    dry_run: bool = False,
    staging_dir: Optional[str] = None,
) -> str:
    """Upload pair texts as JSON. If staging_dir, save locally."""
    hf_path = pair_texts_hf_path(benchmark)
    if dry_run:
        print(f"  [DRY RUN] Would upload {hf_path} ({len(pairs)} pairs)")
        return hf_path

    if staging_dir:
        out_path = Path(staging_dir) / hf_path
        out_path.parent.mkdir(parents=True, exist_ok=True)
        with open(out_path, "w") as f:
            json.dump(pairs, f, indent=JSON_INDENT)
        print(f"  Staged {hf_path} ({len(pairs)} pairs)")
        return hf_path

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as tmp:
        json.dump(pairs, tmp, indent=JSON_INDENT)
        tmp_path = tmp.name
    try:
        api = _get_api()
        api.upload_file(
            path_or_fileobj=tmp_path, path_in_repo=hf_path,
            repo_id=HF_REPO_ID, repo_type=HF_REPO_TYPE,
        )
        print(f"  Uploaded {hf_path} ({len(pairs)} pairs)")
    finally:
        Path(tmp_path).unlink(missing_ok=True)
    return hf_path


def flush_staging_dir(staging_dir: str, path_in_repo: str = "."):
    """Upload everything in staging_dir as a single commit."""
    api = _get_api()
    api.upload_folder(
        folder_path=staging_dir,
        path_in_repo=path_in_repo,
        repo_id=HF_REPO_ID,
        repo_type=HF_REPO_TYPE,
    )
    print(f"  Flushed staging dir to HF ({path_in_repo})")


def write_marker(
    model_name: str,
    benchmark: str,
    strategy: str,
    layers: List[int],
    dry_run: bool = False,
) -> None:
    """Write a per-combo marker file (markers/{safe_model}/{benchmark}/{strategy}.json)."""
    safe_model = model_to_safe_name(model_name)
    marker_path = f"markers/{safe_model}/{benchmark}/{strategy}.json"
    payload = {"layers": sorted(layers)}

    if dry_run:
        print(f"  [DRY RUN] Would write marker: {marker_path} -> {payload}")
        return

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as tmp:
        json.dump(payload, tmp, indent=JSON_INDENT)
        tmp_path = tmp.name
    try:
        api = _get_api()
        api.upload_file(
            path_or_fileobj=tmp_path, path_in_repo=marker_path,
            repo_id=HF_REPO_ID, repo_type=HF_REPO_TYPE,
        )
        print(f"  Wrote marker: {marker_path} -> {payload}")
    finally:
        Path(tmp_path).unlink(missing_ok=True)


def consolidate_index(dry_run: bool = False) -> Dict[str, List[int]]:
    """Build unified index.json from all marker files and upload it."""
    from huggingface_hub import HfApi

    api = HfApi(token=_get_hf_token())
    all_files = api.list_repo_tree(
        repo_id=HF_REPO_ID,
        repo_type=HF_REPO_TYPE,
        path_in_repo="markers",
        recursive=True,
    )
    marker_paths = []
    for f in all_files:
        p = getattr(f, "path", None) or getattr(f, "rpath", None)
        if p and p.endswith(".json"):
            marker_paths.append(p)
    print(f"Found {len(marker_paths)} marker files")
    index: Dict[str, List[int]] = {}
    from huggingface_hub import hf_hub_download
    for mp in marker_paths:
        if not mp.startswith("markers/") or not mp.endswith(".json"):
            continue
        inner = mp[len("markers/"):-len(".json")].split("/")
        if len(inner) < 3:
            continue
        key = f"{inner[0]}/{'/'.join(inner[1:-1])}/{inner[-1]}"
        local = hf_hub_download(
            repo_id=HF_REPO_ID, filename=mp,
            repo_type=HF_REPO_TYPE, token=_get_hf_token(),
        )
        with open(local, "r") as f:
            data = json.load(f)
        existing = set(index.get(key, []))
        existing.update(data.get("layers", []))
        index[key] = sorted(existing)

    print(f"Consolidated index: {len(index)} entries")

    if dry_run:
        for k, v in sorted(index.items()):
            print(f"  {k}: {v}")
        return index

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as tmp:
        json.dump(index, tmp, indent=JSON_INDENT)
        tmp_path = tmp.name
    try:
        api.upload_file(
            path_or_fileobj=tmp_path, path_in_repo="index.json",
            repo_id=HF_REPO_ID, repo_type=HF_REPO_TYPE,
        )
        print(f"  Uploaded consolidated index.json")
    finally:
        Path(tmp_path).unlink(missing_ok=True)

    return index



"""Arabic leaderboard acva group task manifest."""

from __future__ import annotations

BASE_IMPORT = "wisent.extractors.lm_eval.lm_task_extractors."

ARABIC_LEADERBOARD_ACVA_TASKS = {
    "arabic_leaderboard_acva": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Algeria": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Ancient_Egypt": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arab_Empire": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Architecture": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Art": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Astronomy": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Calligraphy": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Ceremony": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Clothing": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Culture": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Food": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Funeral": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Geography": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_History": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Language_Origin": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Literature": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Math": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Medicine": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Music": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Ornament": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Philosophy": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Physics_and_Chemistry": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Arabic_Wedding": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Bahrain": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Comoros": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Egypt_modern": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromAncientEgypt": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromByzantium": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromChina": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromGreece": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromIslam": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromPersia": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_InfluenceFromRome": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Iraq": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Islam_Education": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Islam_branches_and_schools": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Islamic_law_system": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Jordan": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Kuwait": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Lebanon": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Libya": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Mauritania": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Mesopotamia_civilization": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Morocco": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Oman": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Palestine": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Qatar": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Saudi_Arabia": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Somalia": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Sudan": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Syria": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Tunisia": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_United_Arab_Emirates": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_Yemen": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_communication": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_computer_and_phone": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_daily_life": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
    "arabic_leaderboard_acva_entertainment": f"{BASE_IMPORT}arabic_leaderboard_complete:ArabicLeaderboardCompleteExtractor",
}



from __future__ import annotations

import random
from typing import Any, TYPE_CHECKING

from wisent.core.primitives.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.primitives.contrastive_pairs.core.io.response import NegativeResponse, PositiveResponse
from wisent.extractors.lm_eval.atoms import LMEvalBenchmarkExtractor
from wisent.core.utils.cli.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from lm_eval.api.task import ConfigurableTask


__all__ = ["GPQAExtractor"]
_LOG = setup_logger(__name__)

task_names = (
    "gpqa",
    "gpqa_diamond_cot_n_shot",
    "gpqa_diamond_cot_zeroshot",
    "gpqa_diamond_generative_n_shot",
    "gpqa_diamond_n_shot",
    "gpqa_diamond_zeroshot",
    "gpqa_extended_cot_n_shot",
    "gpqa_extended_cot_zeroshot",
    "gpqa_extended_generative_n_shot",
    "gpqa_extended_n_shot",
    "gpqa_extended_zeroshot",
    "gpqa_main_cot_n_shot",
    "gpqa_main_cot_zeroshot",
    "gpqa_main_generative_n_shot",
    "gpqa_main_n_shot",
    "gpqa_main_zeroshot",
)

class GPQAExtractor(LMEvalBenchmarkExtractor):
    """Extractor for GPQA (Graduate-Level Google-Proof Q&A) benchmark."""


    evaluator_name = "generation"
    def extract_contrastive_pairs(
        self,
        lm_eval_task_data: ConfigurableTask,
        limit: int | None = None,
        *,
        train_ratio: float,
    ) -> list[ContrastivePair]:
        """
        Build contrastive pairs from GPQA docs.

        GPQA schema:
            - Question: str
            - Correct Answer: str
            - Incorrect Answer 1-3: str

        Args:
            lm_eval_task_data: lm-eval task instance for GPQA.
            limit: Optional maximum number of pairs to produce.

        Returns:
            A list of ContrastivePair objects.
        """
        log = bind(_LOG, task=getattr(lm_eval_task_data, "NAME", "unknown"))

        max_items = self._normalize_limit(limit)
        docs = self.load_docs(lm_eval_task_data, max_items, train_ratio=train_ratio)

        pairs: list[ContrastivePair] = []

        log.info("Extracting contrastive pairs", extra={"doc_count": len(docs)})

        for doc in docs:
            pair = self._extract_pair_from_doc(doc, lm_eval_task_data)
            if pair is not None:
                pairs.append(pair)
                if max_items is not None and len(pairs) >= max_items:
                    break

        if not pairs:
            task_name = getattr(lm_eval_task_data, "NAME", type(lm_eval_task_data).__name__)
            log.warning("No valid GPQA pairs extracted", extra={"task": task_name})

        return pairs

    def _extract_pair_from_doc(
        self, doc: dict[str, Any], task_data: Any = None
    ) -> ContrastivePair | None:
        """
        Convert a single doc into a ContrastivePair.

        Returns None when required fields are missing or malformed.
        """
        try:
            # Handle both raw format (from Idavidrein/gpqa dataset) and processed format
            if not all(key in doc for key in ["Question", "Correct Answer", "Incorrect Answer 1"]):
                _LOG.debug("Skipping: missing required fields")
                return None

            question = doc["Question"].strip()
            correct_answer = doc["Correct Answer"].strip()
            incorrect_answers = [
                doc.get("Incorrect Answer 1", "").strip(),
                doc.get("Incorrect Answer 2", "").strip(),
                doc.get("Incorrect Answer 3", "").strip(),
            ]
            # Filter out empty answers
            incorrect_answers = [ans for ans in incorrect_answers if ans]

            if not question or not correct_answer or not incorrect_answers:
                _LOG.debug("Skipping: empty question, correct answer, or no incorrect answers")
                return None

            # Create choices list with correct answer first
            choices = [correct_answer] + incorrect_answers

            # Format the question
            if hasattr(task_data, "doc_to_text"):
                formatted_question = task_data.doc_to_text(doc)
            else:
                choice_letters = ["A", "B", "C", "D"]
                formatted_choices = []
                for i, choice in enumerate(choices[:4]):
                    if choice:
                        formatted_choices.append(f"({choice_letters[i]}) {choice}")
                formatted_question = f"{question}\nChoices:\n" + "\n".join(formatted_choices)

            # Randomly select one incorrect answer
            incorrect_answer = random.choice(incorrect_answers)

            metadata = {
                "label": "gpqa",
                "source": getattr(task_data, "NAME", "gpqa"),
            }

            return self._build_pair(
                question=formatted_question,
                correct=correct_answer,
                incorrect=incorrect_answer,
                metadata=metadata,
            )

        except Exception as exc:
            _LOG.error(f"Error extracting pair from doc: {exc}", exc_info=True)
            return None

    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        """Build a ContrastivePair from question and responses."""
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(
            prompt=question,
            positive_response=positive_response,
            negative_response=negative_response,
            label=metadata.get("label") if metadata else None,
            metadata=metadata,
        )



"""
Attention-Transport steering method (SZLAK).

Uses attention-affinity cost matrix C_ij = -q_i·k_j/√d_k and one-sided
entropic optimal transport to compute per-source displacements.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple
from dataclasses import dataclass, field
from collections import defaultdict

import torch
import torch.nn.functional as F

from wisent.core.control.steering_methods.core.atoms import BaseSteeringMethod
from wisent.core.primitives.model_interface.core.activations.core.atoms import LayerActivations, RawActivationMap, LayerName
from wisent.core.primitives.contrastive_pairs.core.set import ContrastivePairSet
from wisent.core.utils.infra_tools.errors import InsufficientDataError
from wisent.core.utils.config_tools.constants import LOG_EPS

from .transport import (
    compute_attention_affinity_cost,
    sinkhorn_one_sided,
)

__all__ = [
    "SzlakMethod",
    "SzlakConfig",
    "SzlakResult",
]


def _require(name: str, kwargs: dict):
    """Raise ValueError if a required hyperparameter is missing."""
    if name not in kwargs:
        raise ValueError(
            f"Parameter '{name}' is required. "
            f"Run 'wisent optimize-steering auto' first, or pass it explicitly."
        )
    return kwargs[name]


@dataclass
class SzlakConfig:
    """Configuration for Attention-Transport steering method."""
    sinkhorn_reg: float
    """Entropic regularization for one-sided EOT solver."""
    max_iter: int
    """Maximum Sinkhorn iterations (unused for one-sided, kept for API consistency)."""
    inference_k: int
    """Number of nearest source points for inference interpolation."""


@dataclass
class SzlakResult:
    """Result from Geodesic OT training with per-layer transport data."""
    source_points: Dict[LayerName, torch.Tensor]
    """Per-layer source (negative) activations [N_neg, D]."""
    displacements: Dict[LayerName, torch.Tensor]
    """Per-layer transport displacements [N_neg, D]."""
    mean_displacement: Dict[LayerName, torch.Tensor]
    """Per-layer mean displacement vector [D]."""
    config: SzlakConfig
    layer_order: List[LayerName]
    metadata: Dict[str, Any] = field(default_factory=dict)


class SzlakMethod(BaseSteeringMethod):
    """
    Attention-Transport steering method.

    Computes attention-affinity cost C = -(q@k.T)/√d_k and solves
    one-sided entropic OT to derive displacements.
    """

    name = "szlak"
    description = (
        "Attention-transport steering via EOT cost inversion"
    )

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.config = SzlakConfig(
            sinkhorn_reg=_require("sinkhorn_reg", kwargs),
            max_iter=_require("max_iter", kwargs),
            inference_k=_require("inference_k", kwargs),
        )

    def train(self, pair_set: ContrastivePairSet) -> LayerActivations:
        """Compatibility interface: returns mean transport vectors per layer."""
        result = self.train_szlak(pair_set)
        primary_map: RawActivationMap = {}
        for layer in result.layer_order:
            vec = result.mean_displacement[layer]
            vec = F.normalize(vec.float(), p=2, dim=-1)
            primary_map[layer] = vec
        dtype = self.kwargs.get("dtype", None)
        return LayerActivations(primary_map, dtype=dtype)

    def train_szlak(
        self,
        pair_set: ContrastivePairSet,
        qk_activations: Dict[LayerName, Tuple[torch.Tensor, torch.Tensor]] | None = None,
        num_heads: int | None = None,
        num_kv_heads: int | None = None,
    ) -> SzlakResult:
        """Full attention-transport training with per-layer displacements.

        Args:
            pair_set: Contrastive pair set with activations.
            qk_activations: Q/K projections per layer (layer_name -> (q_neg, k_pos)).
            num_heads: Number of attention heads (for GQA handling).
            num_kv_heads: Number of key-value heads (for GQA handling).
        """
        buckets = self._collect_from_set(pair_set)
        if not buckets:
            raise InsufficientDataError(reason="No valid activation pairs found")
        if not qk_activations:
            raise InsufficientDataError(reason="Q/K projections required for attention-transport")
        source_points: Dict[LayerName, torch.Tensor] = {}
        displacements_dict: Dict[LayerName, torch.Tensor] = {}
        mean_disp: Dict[LayerName, torch.Tensor] = {}
        layer_order: List[LayerName] = []
        layer_meta: Dict[str, Any] = {}
        for layer_name in sorted(buckets.keys()):
            pos_list, neg_list = buckets[layer_name]
            if not pos_list or not neg_list:
                continue
            if layer_name not in qk_activations:
                continue
            pos = torch.stack([t.detach().float().reshape(-1) for t in pos_list], dim=0)
            neg = torch.stack([t.detach().float().reshape(-1) for t in neg_list], dim=0)
            q_neg, k_pos = qk_activations[layer_name]
            cost = compute_attention_affinity_cost(
                q_neg.float(), k_pos.float(), num_heads=num_heads, num_kv_heads=num_kv_heads)
            T = sinkhorn_one_sided(cost, reg=self.config.sinkhorn_reg, max_iter=self.config.max_iter)
            row_sums = T.sum(dim=1, keepdim=True).clamp(min=LOG_EPS)
            T_norm = T / row_sums
            targets = T_norm @ pos
            delta = targets - neg
            source_points[layer_name] = neg.detach()
            displacements_dict[layer_name] = delta.detach()
            mean_disp[layer_name] = delta.mean(dim=0).detach()
            layer_order.append(layer_name)
            layer_meta[str(layer_name)] = {
                "n_neg": neg.shape[0], "n_pos": pos.shape[0],
                "mean_displacement_norm": delta.mean(dim=0).norm().item(),
                "max_cost": cost.max().item(),
            }
        if not layer_order:
            raise InsufficientDataError(reason="No layers produced valid transport")
        return SzlakResult(
            source_points=source_points,
            displacements=displacements_dict,
            mean_displacement=mean_disp,
            config=self.config,
            layer_order=layer_order,
            metadata={"per_layer": layer_meta},
        )

    def _collect_from_set(
        self, pair_set: ContrastivePairSet
    ) -> Dict[LayerName, Tuple[List[torch.Tensor], List[torch.Tensor]]]:
        """Build {layer_name: ([pos tensors], [neg tensors])} from pairs."""
        buckets: Dict[
            LayerName, Tuple[List[torch.Tensor], List[torch.Tensor]]
        ] = defaultdict(lambda: ([], []))
        for pair in pair_set.pairs:
            pos_la = getattr(pair.positive_response, "layers_activations", None)
            neg_la = getattr(pair.negative_response, "layers_activations", None)
            if pos_la is None or neg_la is None:
                continue
            layer_names = set(pos_la.to_dict().keys()) | set(neg_la.to_dict().keys())
            for layer in layer_names:
                p = pos_la.to_dict().get(layer) if pos_la else None
                n = neg_la.to_dict().get(layer) if neg_la else None
                if isinstance(p, torch.Tensor) and isinstance(n, torch.Tensor):
                    buckets[layer][0].append(p)
                    buckets[layer][1].append(n)
        return buckets



"""Runtime hardware resource detection for dynamic configuration."""
from __future__ import annotations

import os
import re
from dataclasses import dataclass
from functools import lru_cache

from wisent.core.utils.config_tools.constants import (
    FP16_GB_PER_BILLION_PARAMS,
    FP16_BYTES,
    KV_PAIR_FACTOR,
    CUDA_CONTEXT_MB,
    GPU_FRAMEWORK_OVERHEAD_GB,
    MB_PER_GB,
    COMBO_OFFSET,
    TRANSFORMER_PARAM_FACTOR,
    PARAMS_PER_BILLION,
    GENERATION_DEFAULT_MAX_NEW_TOKENS,
)
from wisent.core.utils.config_tools.constants.validated._validated import BYTES_PER_MB


@dataclass(frozen=True)
class SystemResources:
    """Detected system hardware resources."""

    cpu_count: int
    total_ram_mb: int
    gpu_mem_mb: int  # 0 if no GPU


@lru_cache(maxsize=1)
def detect_system_resources() -> SystemResources:
    """Detect CPU count, total RAM, and GPU memory."""
    cpu_count = os.cpu_count() or 4
    try:
        pages = os.sysconf("SC_PHYS_PAGES")
        page_size = os.sysconf("SC_PAGE_SIZE")
        total_ram_mb = (pages * page_size) // (1024 * 1024)
    except (ValueError, OSError, AttributeError):
        total_ram_mb = 8192
    gpu_mem_mb = 0
    try:
        import torch
        if torch.cuda.is_available():
            gpu_mem_mb = (
                torch.cuda.get_device_properties(0).total_memory
                // (1024 * 1024)
            )
    except Exception:
        pass
    return SystemResources(
        cpu_count=cpu_count,
        total_ram_mb=total_ram_mb,
        gpu_mem_mb=gpu_mem_mb,
    )


# ---------------------------------------------------------------------------
# Docker limits -- scale with available RAM and CPU
# ---------------------------------------------------------------------------

def docker_mem_limit_mb() -> int:
    """25 pct of total RAM, clamped [512, 16384]."""
    res = detect_system_resources()
    return max(512, min(16384, res.total_ram_mb // 4))


def docker_nproc() -> int:
    """32 x cpu_count, clamped [64, 1024]."""
    res = detect_system_resources()
    return max(64, min(1024, 32 * res.cpu_count))


def docker_nofile() -> int:
    """4 x nproc."""
    return 4 * docker_nproc()


def docker_cpu_limit_s() -> int:
    """max(2, cpu_count // 2)."""
    res = detect_system_resources()
    return max(2, res.cpu_count // 2)


def docker_wall_timeout_s() -> int:
    """3 x cpu_limit."""
    return 3 * docker_cpu_limit_s()


def docker_fsize_mb() -> int:
    """Fixed filesystem limit."""
    return 16


def docker_pids_limit() -> int:
    """2 x nproc."""
    return 2 * docker_nproc()


# ---------------------------------------------------------------------------
# Eval limits -- derived from Docker limits
# ---------------------------------------------------------------------------

def eval_mem_limit_mb() -> int:
    """docker_mem_limit_mb() // 4, min 256."""
    return max(256, docker_mem_limit_mb() // 4)


def eval_cpu_limit_s() -> int:
    """Same as docker_cpu_limit_s."""
    return docker_cpu_limit_s()


def eval_time_limit_s() -> int:
    """Same as docker_wall_timeout_s."""
    return docker_wall_timeout_s()


# ---------------------------------------------------------------------------
# Batch sizes -- scale with GPU / CPU memory
# ---------------------------------------------------------------------------

def default_batch_size() -> int:
    """GPU: gpu_mem // 3072, clamped [1, 32]. No GPU: ram // 4096, [1, 8]."""
    res = detect_system_resources()
    if res.gpu_mem_mb > 0:
        return max(1, min(32, res.gpu_mem_mb // 3072))
    return max(1, min(8, res.total_ram_mb // 4096))


def eval_batch_size() -> int:
    """default_batch_size() // 2, min 1."""
    return max(1, default_batch_size() // 2)


def extraction_batch_size() -> int:
    """default_batch_size() // 2, min 1."""
    return max(1, default_batch_size() // 2)


# ---------------------------------------------------------------------------
# Subprocess timeouts -- scale inversely with CPU count
# ---------------------------------------------------------------------------

def subprocess_timeout_s() -> int:
    """max(60, 600 // max(1, cpu_count // 4))."""
    res = detect_system_resources()
    return max(60, 600 // max(1, res.cpu_count // 4))


def subprocess_timeout_long_s() -> int:
    """2 x subprocess_timeout_s, min 300."""
    return max(300, 2 * subprocess_timeout_s())


# ---------------------------------------------------------------------------
# Sandbox-specific limits
# ---------------------------------------------------------------------------

def docker_bigcode_mem_limit_mb() -> int:
    """docker_mem_limit_mb() // 16, clamped [128, 1024]."""
    return max(128, min(1024, docker_mem_limit_mb() // 16))


def docker_code_exec_timeout_s() -> int:
    """5 x docker_wall_timeout_s, min 60."""
    return max(60, 5 * docker_wall_timeout_s())


def code_eval_mem_limit_mb() -> int:
    """Same as eval_mem_limit_mb."""
    return eval_mem_limit_mb()


def docker_sandbox_time_limit_s() -> int:
    """docker_wall_timeout_s // 2, min 5."""
    return max(5, docker_wall_timeout_s() // 2)


def docker_sandbox_cpu_limit_s() -> int:
    """docker_cpu_limit_s // 2, min 2."""
    return max(2, docker_cpu_limit_s() // 2)


def docker_sandbox_mem_limit_mb() -> int:
    """docker_mem_limit_mb() // 8, clamped [256, 2048]."""
    return max(256, min(2048, docker_mem_limit_mb() // 8))


def safe_docker_nproc_default() -> int:
    """docker_nproc() // 2, min 32."""
    return max(32, docker_nproc() // 2)


# ---------------------------------------------------------------------------
# GPU memory estimation -- per-worker and max-parallel-workers
# ---------------------------------------------------------------------------


def _extract_param_billions(model_name: str) -> float:
    """Extract parameter count in billions from model name."""
    match = re.search(r'(\d+\.?\d*)[Bb]', model_name)
    if match:
        return float(match.group(COMBO_OFFSET))
    return _estimate_params_from_config(model_name)


def _estimate_params_from_config(model_name: str) -> float:
    """Estimate parameter count in billions from HuggingFace AutoConfig."""
    from transformers import AutoConfig
    cfg = AutoConfig.from_pretrained(model_name, trust_remote_code=True)
    hidden = getattr(cfg, "hidden_size", None)
    layers = getattr(cfg, "num_hidden_layers", None)
    vocab = getattr(cfg, "vocab_size", None)
    if hidden and layers and vocab:
        total_params = (
            TRANSFORMER_PARAM_FACTOR * layers * hidden * hidden
            + vocab * hidden
        )
        return total_params / PARAMS_PER_BILLION
    raise ValueError(f"Cannot estimate parameters for {model_name}")


def _kv_cache_mb(model_name: str) -> int:
    """KV cache MB at max generation length from model config."""
    from transformers import AutoConfig
    cfg = AutoConfig.from_pretrained(model_name, trust_remote_code=True)
    n_layers = getattr(cfg, "num_hidden_layers", None)
    n_kv = getattr(cfg, "num_key_value_heads", None) or getattr(cfg, "num_attention_heads", None)
    n_heads = getattr(cfg, "num_attention_heads", None)
    h_dim = getattr(cfg, "head_dim", None)
    if h_dim is None and getattr(cfg, "hidden_size", None) and n_heads:
        h_dim = cfg.hidden_size // n_heads
    if not all((n_layers, n_kv, h_dim)):
        return GPU_FRAMEWORK_OVERHEAD_GB * MB_PER_GB
    return n_layers * KV_PAIR_FACTOR * n_kv * h_dim * GENERATION_DEFAULT_MAX_NEW_TOKENS * FP16_BYTES // BYTES_PER_MB


def estimate_model_memory_mb(model_name: str) -> int:
    """Estimate per-worker GPU memory in MB: weights + KV cache + CUDA context."""
    params_b = _extract_param_billions(model_name)
    weights_mb = int(params_b * FP16_GB_PER_BILLION_PARAMS * MB_PER_GB)
    return weights_mb + _kv_cache_mb(model_name) + CUDA_CONTEXT_MB


def estimate_max_gpu_workers(model_name: str) -> int:
    """Estimate max parallel workers fitting on the current GPU."""
    res = detect_system_resources()
    if not res.gpu_mem_mb:
        return COMBO_OFFSET
    per_worker_mb = estimate_model_memory_mb(model_name)
    workers = res.gpu_mem_mb // per_worker_mb
    return max(COMBO_OFFSET, workers)


# ---------------------------------------------------------------------------
# DS1000 overrides -- scaled from base limits
# ---------------------------------------------------------------------------

def ds1000_cpu_limit_s() -> int:
    """10 x docker_cpu_limit_s, min 30."""
    return max(30, 10 * docker_cpu_limit_s())


def ds1000_wall_timeout_s() -> int:
    """2 x ds1000_cpu_limit_s."""
    return 2 * ds1000_cpu_limit_s()


def ds1000_nproc() -> int:
    """4 x docker_nproc, clamped [256, 2048]."""
    return max(256, min(2048, 4 * docker_nproc()))



"""Okapi mmlu multilingual group task manifest."""

from __future__ import annotations

BASE_IMPORT = "wisent.extractors.lm_eval.lm_task_extractors."

# Individual mmlu tasks
_MMLU_TASKS = [
    "m_mmlu_ar", "m_mmlu_bn", "m_mmlu_ca", "m_mmlu_da", "m_mmlu_de", "m_mmlu_en",
    "m_mmlu_es", "m_mmlu_eu", "m_mmlu_fr", "m_mmlu_gu", "m_mmlu_hi", "m_mmlu_hr",
    "m_mmlu_hu", "m_mmlu_hy", "m_mmlu_id", "m_mmlu_is", "m_mmlu_it", "m_mmlu_kn",
    "m_mmlu_ml", "m_mmlu_mr", "m_mmlu_nb", "m_mmlu_ne", "m_mmlu_nl", "m_mmlu_pt",
    "m_mmlu_ro", "m_mmlu_ru", "m_mmlu_sk", "m_mmlu_sr", "m_mmlu_sv", "m_mmlu_ta",
    "m_mmlu_te", "m_mmlu_uk", "m_mmlu_vi", "m_mmlu_zh"
]

OKAPI_MMLU_MULTILINGUAL_TASKS = {
    "okapi_mmlu_multilingual": f"{BASE_IMPORT}okapi_mmlu_multilingual:OkapiMmluMultilingualExtractor",
    "okapi/mmlu_multilingual": f"{BASE_IMPORT}okapi_mmlu_multilingual:OkapiMmluMultilingualExtractor",
}

# Register all individual tasks
for task in _MMLU_TASKS:
    OKAPI_MMLU_MULTILINGUAL_TASKS[task] = f"{BASE_IMPORT}okapi_mmlu_multilingual:OkapiMmluMultilingualExtractor"



"""Human-readable benchmark descriptions for matching."""

from __future__ import annotations


__all__ = ["BENCHMARK_DESCRIPTIONS"]


BENCHMARK_DESCRIPTIONS = {
    # Knowledge & QA
    "mmlu": "General knowledge across academic subjects (science, history, literature, etc.)",
    "triviaqa": "Trivia and factual questions from various domains",
    "naturalqs": "Natural questions asking for factual information",
    "webqs": "Web-based questions requiring factual knowledge",
    "arc_easy": "Elementary science questions and reasoning",
    "arc_challenge": "Advanced science questions and reasoning",
    "sciq": "Scientific questions and explanations",
    "social_iqa": "Social situations and common sense reasoning",
    "openbookqa": "Elementary science with open-book style questions",
    "gpqa": "Graduate-level scientific reasoning in biology, physics, and chemistry",
    "gpqa_diamond": "High-quality graduate-level scientific questions (premium subset)",
    "gpqa_extended": "Extended graduate-level scientific reasoning dataset",
    "gpqa_main_zeroshot": "Graduate-level scientific reasoning (main subset, zero-shot)",
    "gpqa_diamond_zeroshot": "Premium graduate-level scientific questions (zero-shot)",
    "gpqa_extended_zeroshot": "Extended graduate-level scientific reasoning (zero-shot)",
    "gpqa_main_cot_zeroshot": "Graduate-level scientific reasoning with chain-of-thought",
    "gpqa_diamond_cot_zeroshot": "Premium graduate-level scientific questions with reasoning",
    "gpqa_extended_cot_zeroshot": "Extended graduate-level scientific reasoning with CoT",
    # SuperGPQA
    "supergpqa": "Large-scale dataset of scientific multiple-choice questions across disciplines",
    "supergpqa_physics": "Large-scale physics multiple-choice questions",
    "supergpqa_chemistry": "Large-scale chemistry multiple-choice questions",
    "supergpqa_biology": "Large-scale biology multiple-choice questions",
    # HLE
    "hle": "Human-Level Evaluation: Multimodal reasoning across multiple domains",
    "hle_exact_match": "Human-Level Evaluation: Exact string matching questions",
    "hle_multiple_choice": "Human-Level Evaluation: Multiple choice questions",
    # Reading Comprehension
    "coqa": "Conversational question answering with context",
    "drop": "Reading comprehension with numerical reasoning",
    "race": "Reading comprehension from English exams",
    "squad2": "Reading comprehension with impossible questions",
    "qasper": "Scientific paper question answering",
    "mutual": "Dialogue understanding and reasoning",
    # Reasoning & Logic
    "hellaswag": "Commonsense reasoning about everyday situations",
    "piqa": "Physical reasoning about objects and actions",
    "winogrande": "Pronoun resolution requiring commonsense",
    "logiqa": "Logical reasoning and inference",
    "wsc273": "Winograd schema challenge for pronoun resolution",
    "swag": "Commonsense reasoning about video situations",
    "boolq": "Yes/no questions requiring reasoning",
    # Mathematics
    "gsm8k": "Grade school math word problems",
    "math": "Mathematical reasoning problems requiring multi-step solutions",
    "math500": "500-problem subset of MATH benchmark for mathematical reasoning",
    "hendrycks_math": "Competition-level mathematics problems from Hendrycks et al.",
    "aime": "High-difficulty AIME contest problems (latest: 2025)",
    "aime2025": "High-difficulty AIME contest problems from 2025 (MathArena)",
    "aime2024": "High-difficulty AIME contest problems from 2024",
    "hmmt": "High-difficulty HMMT contest problems (latest: February 2025)",
    "hmmt_feb_2025": "High-difficulty HMMT February 2025 contest problems",
    "polymath": "PolyMath multilingual mathematical reasoning (default: English medium)",
    "polymath_en_medium": "PolyMath medium-difficulty mathematical problems in English",
    "polymath_zh_medium": "PolyMath medium-difficulty mathematical problems in Chinese",
    "polymath_en_high": "PolyMath high-difficulty mathematical problems in English",
    "polymath_zh_high": "PolyMath high-difficulty mathematical problems in Chinese",
    "livemathbench": "LiveMathBench CNMO 2024 mathematical olympiad problems",
    "livemathbench_cnmo_en": "LiveMathBench CNMO 2024 problems in English",
    "livemathbench_cnmo_zh": "LiveMathBench CNMO 2024 problems in Chinese",
    "math_qa": "Mathematical reasoning and problem solving",
    "arithmetic": "Basic arithmetic operations and calculations",
    "asdiv": "Arithmetic story problems for children",
    "mgsm": "Multilingual grade school math problems",
    # Coding
    "humaneval": "Python code generation and programming",
    "mbpp": "Python programming problems and solutions",
    # Language & Linguistics
    "blimp": "Grammatical acceptability and linguistic knowledge",
    "lambada": "Language modeling and word prediction",
    "lambada_cloze": "Cloze test for language understanding",
    "lambada_multilingual": "Multilingual language modeling",
    "wikitext": "Language modeling on Wikipedia text",
    "unscramble": "Word unscrambling and letter manipulation",
    # Multilingual
    "xnli": "Cross-lingual natural language inference",
    "xcopa": "Cross-lingual commonsense reasoning",
    "xstorycloze": "Cross-lingual story completion",
    "xwinograd": "Cross-lingual pronoun resolution",
    "paws_x": "Cross-lingual paraphrase detection",
    "belebele": "Multilingual reading comprehension",
    # Bias & Safety
    "toxigen": "Toxicity detection and harmful content",
    "crows_pairs": "Bias measurement in language models",
    "hendrycks_ethics": "Ethical reasoning and moral judgments",
    "truthfulqa_mc1": "Truthfulness and factual accuracy",
    "truthfulqa_mc2": "Truthfulness with multiple correct answers",
    "truthfulqa_gen": "Truthful text generation",
    # Medical
    "medqa": "Medical knowledge and clinical reasoning",
    "pubmedqa": "Biomedical literature question answering",
    "headqa": "Medical and healthcare knowledge",
    # Temporal & Event
    "prost": "Temporal reasoning in procedural text",
    # Adversarial
    "anli": "Adversarial natural language inference",
    # Benchmark Suites
    "glue": "General language understanding tasks",
    "superglue": "Advanced language understanding tasks",
    "big_bench": "Diverse challenging tasks for large models",
}



"""TECZA and TETNO steering object creation helpers."""
from __future__ import annotations
import torch

from wisent.core import constants as _C
from wisent.core.control.steering_methods.configs.optimal import get_optimal


def _require_arg(args, attr_name):
    """Get required arg or raise clear error."""
    val = getattr(args, attr_name, None)
    if val is None:
        raise ValueError(
            f"Parameter '{attr_name}' is required. "
            f"Run 'wisent optimize-steering auto' first, or pass it explicitly."
        )
    return val


def _create_tecza_steering_object(
    metadata: SteeringObjectMetadata,
    layer_activations: dict,
    available_layers: list,
    args,
) -> TECZASteeringObject:
    """Create TECZA steering object with multiple directions."""
    from wisent.core.control.steering_methods.methods.advanced import TECZAMethod

    _r = _require_arg
    method = TECZAMethod(
        num_directions=_r(args, 'tecza_num_directions'),
        optimization_steps=_r(args, 'tecza_optimization_steps'),
        learning_rate=_r(args, 'tecza_learning_rate'),
        retain_weight=_r(args, 'tecza_retain_weight'),
        independence_weight=_r(args, 'tecza_independence_weight'),
        min_cosine_similarity=_r(args, 'tecza_min_cosine_sim'),
        max_cosine_similarity=_r(args, 'tecza_max_cosine_sim'),
        variance_threshold=_r(args, 'tecza_variance_threshold'),
        marginal_threshold=_r(args, 'tecza_marginal_threshold'),
        max_directions=_r(args, 'tecza_max_directions'),
        ablation_weight=_r(args, 'tecza_ablation_weight'),
        addition_weight=_r(args, 'tecza_addition_weight'),
        separation_margin=_r(args, 'tecza_separation_margin'),
        perturbation_scale=_r(args, 'tecza_perturbation_scale'),
        universal_basis_noise=_r(args, 'tecza_universal_basis_noise'),
        log_interval=_r(args, 'tecza_log_interval'),
        normalize=getattr(args, 'normalize', get_optimal("normalize")),
    )
    
    directions = {}
    direction_weights = {}
    
    for layer_str in available_layers:
        pos_list = layer_activations[layer_str]["positive"]
        neg_list = layer_activations[layer_str]["negative"]
        
        if not pos_list or not neg_list:
            continue
        
        # Stack activations
        pos_tensor = torch.stack([t.detach().float().reshape(-1) for t in pos_list], dim=0)
        neg_tensor = torch.stack([t.detach().float().reshape(-1) for t in neg_list], dim=0)
        
        # Train directions
        layer_dirs, meta = method._train_layer_directions(pos_tensor, neg_tensor, layer_str, log_interval=method.log_interval)
        
        layer_int = int(layer_str)
        directions[layer_int] = layer_dirs
        # Equal weights by default
        direction_weights[layer_int] = torch.ones(layer_dirs.shape[0]) / layer_dirs.shape[0]
        
        print(f"   Layer {layer_str}: {layer_dirs.shape[0]} directions, avg_cosine={meta['avg_cosine_similarity']:.3f}")
    
    from wisent.core.control.steering_methods._steering_object_advanced import TECZASteeringObject
    return TECZASteeringObject(
        metadata=metadata,
        directions=directions,
        direction_weights=direction_weights,
        primary_only=False,
    )


def _create_tetno_steering_object(
    metadata: SteeringObjectMetadata,
    layer_activations: dict,
    available_layers: list,
    args,
    *,
    threshold_search_init_value: float,
    default_score: float,
) -> TETNOSteeringObject:
    """Create TETNO steering object with conditional gating."""
    # Determine sensor layer — use last available layer if not specified
    num_layers = len(available_layers)
    sensor_layer_idx = getattr(args, 'tetno_sensor_layer', None)
    if sensor_layer_idx is None:
        sensor_layer_idx = num_layers - _C.COMBO_OFFSET
    sensor_layer = int(available_layers[min(sensor_layer_idx, num_layers - _C.COMBO_OFFSET)])
    
    # Train behavior vectors
    behavior_vectors = {}
    layer_scales = {}
    
    for layer_str in available_layers:
        pos_list = layer_activations[layer_str]["positive"]
        neg_list = layer_activations[layer_str]["negative"]
        
        if not pos_list or not neg_list:
            continue
        
        pos_tensor = torch.stack([t.detach().float().reshape(-1) for t in pos_list], dim=0)
        neg_tensor = torch.stack([t.detach().float().reshape(-1) for t in neg_list], dim=0)
        
        # CAA for behavior vector
        behavior_vec = pos_tensor.mean(dim=0) - neg_tensor.mean(dim=0)
        behavior_vec = torch.nn.functional.normalize(behavior_vec, dim=0)
        
        layer_int = int(layer_str)
        behavior_vectors[layer_int] = behavior_vec
        
        # Compute layer scale based on separation
        pos_proj = (pos_tensor * behavior_vec).sum(dim=1).mean()
        neg_proj = (neg_tensor * behavior_vec).sum(dim=1).mean()
        separation = (pos_proj - neg_proj).item()
        layer_scales[layer_int] = max(_require_arg(args, 'tetno_min_layer_scale'), separation)
        
        print(f"   Layer {layer_str}: separation={separation:.3f}")
    
    # Normalize scales
    if layer_scales:
        mean_scale = sum(layer_scales.values()) / len(layer_scales)
        if mean_scale > 0:
            layer_scales = {k: v / mean_scale for k, v in layer_scales.items()}
    
    # Train condition vector at sensor layer
    sensor_layer_str = str(sensor_layer)
    pos_sensor = torch.stack([t.detach().float().reshape(-1) for t in layer_activations[sensor_layer_str]["positive"]], dim=0)
    neg_sensor = torch.stack([t.detach().float().reshape(-1) for t in layer_activations[sensor_layer_str]["negative"]], dim=0)
    
    # Initialize with CAA direction
    condition_vec = pos_sensor.mean(dim=0) - neg_sensor.mean(dim=0)
    condition_vec = torch.nn.functional.normalize(condition_vec, dim=0)
    
    # Optimize condition vector
    condition_vec = condition_vec.clone().requires_grad_(True)
    optimizer = torch.optim.Adam([condition_vec], lr=_require_arg(args, 'tetno_learning_rate'))
    
    for _ in range(_require_arg(args, 'tetno_optimization_steps')):
        optimizer.zero_grad()
        c_norm = torch.nn.functional.normalize(condition_vec, dim=0)
        pos_norm = torch.nn.functional.normalize(pos_sensor, dim=1)
        neg_norm = torch.nn.functional.normalize(neg_sensor, dim=1)
        
        pos_sim = (pos_norm * c_norm).sum(dim=1)
        neg_sim = (neg_norm * c_norm).sum(dim=1)
        
        margin = _require_arg(args, 'tetno_condition_margin')
        loss = torch.relu(margin - pos_sim).mean() + torch.relu(neg_sim + margin).mean()
        loss.backward()
        optimizer.step()
    
    condition_vec = torch.nn.functional.normalize(condition_vec.detach(), dim=0)
    
    # Find optimal threshold
    pos_norm = torch.nn.functional.normalize(pos_sensor, dim=1)
    neg_norm = torch.nn.functional.normalize(neg_sensor, dim=1)
    pos_sims = (pos_norm * condition_vec).sum(dim=1)
    neg_sims = (neg_norm * condition_vec).sum(dim=1)
    
    best_threshold, best_acc = threshold_search_init_value, default_score
    threshold_search_steps = _require_arg(args, 'tetno_threshold_search_steps')
    for t in torch.linspace(pos_sims.min(), pos_sims.max(), threshold_search_steps):
        acc = ((pos_sims > t).float().mean() + (neg_sims <= t).float().mean()) / 2
        if acc > best_acc:
            best_acc, best_threshold = acc.item(), t.item()
    
    print(f"   Sensor layer: {sensor_layer}")
    print(f"   Optimal threshold: {best_threshold:.3f} (accuracy: {best_acc:.3f})")
    
    from wisent.core.control.steering_methods._steering_object_advanced import TETNOSteeringObject
    return TETNOSteeringObject(
        metadata=metadata,
        behavior_vectors=behavior_vectors,
        condition_vector=condition_vec,
        sensor_layer=sensor_layer,
        threshold=best_threshold,
        layer_scales=layer_scales,
        gate_temperature=_require_arg(args, 'tetno_gate_temperature'),
    )





"""
Multi-direction weight modification: Bake GROM/TECZA/TETNO directions into weights.

These methods learn multiple steering directions per layer. When baking into weights,
we lose the input-dependent gating but keep the multi-directional coverage.

Baking formula:
    W_new = W + S(weight_i * direction_i)

Where weights are either:
1. Learned static weights from the method
2. Average activation weights across training data
3. Uniform weights (simple average)
"""

from __future__ import annotations

import torch
from typing import Dict, List, Optional, Any, TYPE_CHECKING
from dataclasses import dataclass

from wisent.core.weight_modification.methods.additive import bake_steering_into_weights
from wisent.core.utils.cli.cli_logger import setup_logger, bind
from wisent.core.utils.cli.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from torch import Tensor
    from torch.nn import Module

__all__ = [
    "MultiDirectionConfig",
    "MultiDirectionResult",
    "train_and_bake_grom",
    "train_and_bake_tecza",
    "train_and_bake_tetno",
    "bake_multi_directions",
    "combine_directions",
]

_LOG = setup_logger(__name__)


@dataclass
class MultiDirectionConfig:
    """Configuration for multi-direction weight modification."""

    alpha: float
    """Global steering strength multiplier"""

    num_directions: int
    """Number of directions per layer"""

    optimization_steps: int
    """Training steps for direction optimization"""

    method: Optional[str] = None
    """Which method to use: 'grom', 'tecza', or 'tetno'"""

    combination_strategy: Optional[str] = None
    """How to combine directions: 'learned', 'uniform', 'pca_weighted'"""

    components: List[str] = None
    """Components to modify. Default: ['self_attn.o_proj', 'mlp.down_proj']"""

    bake_method: Optional[str] = None
    """How to bake: 'bias' or 'weight'"""

    def __post_init__(self):
        if self.components is None:
            self.components = ["self_attn.o_proj", "mlp.down_proj"]


@dataclass
class MultiDirectionResult:
    """Result from multi-direction training and baking."""

    method: str
    num_directions: int
    layers_modified: int
    directions_per_layer: Dict[str, int]
    combination_weights: Dict[str, List[float]]
    effective_vectors: Dict[int, "Tensor"]
    metadata: Dict[str, Any]


def combine_directions(
    directions: "Tensor",
    weights: Optional["Tensor"] = None,
    *,
    strategy: str,
) -> "Tensor":
    """
    Combine multiple directions into a single effective direction.

    Args:
        directions: [num_directions, hidden_dim] tensor
        weights: Optional [num_directions] weights
        strategy: 'learned', 'uniform', or 'pca_weighted'

    Returns:
        [hidden_dim] combined direction
    """
    num_dirs = directions.shape[0]

    if strategy == "uniform":
        return directions.mean(dim=0)

    elif strategy == "learned":
        if weights is None:
            weights = torch.ones(num_dirs) / num_dirs
        else:
            weights = torch.softmax(weights, dim=0)
        return (weights.unsqueeze(1) * directions).sum(dim=0)

    elif strategy == "pca_weighted":
        variance_weights = torch.tensor(
            [1.0 / (i + 1) for i in range(num_dirs)])
        variance_weights = variance_weights / variance_weights.sum()
        return (variance_weights.unsqueeze(1).to(directions.device)
                * directions).sum(dim=0)

    else:
        raise ValueError(f"Unknown strategy: {strategy}")


def bake_multi_directions(
    model: "Module",
    directions: Dict[str, "Tensor"],
    weights: Optional[Dict[str, "Tensor"]] = None,
    config: Optional[MultiDirectionConfig] = None,
) -> MultiDirectionResult:
    """Bake multi-direction steering into model weights."""
    cfg = config or MultiDirectionConfig()
    log = bind(_LOG, method=cfg.method)

    effective_vectors: Dict[int, "Tensor"] = {}
    combination_weights: Dict[str, List[float]] = {}
    directions_per_layer: Dict[str, int] = {}

    for layer_name, layer_dirs in directions.items():
        try:
            layer_idx = int(layer_name.replace("layer_", ""))
        except ValueError:
            layer_idx = int(layer_name)

        layer_weights = weights.get(layer_name) if weights else None

        effective = combine_directions(
            layer_dirs, layer_weights, strategy=cfg.combination_strategy)

        effective_vectors[layer_idx] = effective
        directions_per_layer[layer_name] = layer_dirs.shape[0]

        if layer_weights is not None:
            combination_weights[layer_name] = (
                torch.softmax(layer_weights, dim=0).tolist())
        else:
            num_dirs = layer_dirs.shape[0]
            combination_weights[layer_name] = [1.0 / num_dirs] * num_dirs

    stats = bake_steering_into_weights(
        model=model, steering_vectors=effective_vectors,
        components=cfg.components, alpha=cfg.alpha,
        method=cfg.bake_method, verbose=True)

    return MultiDirectionResult(
        method=cfg.method, num_directions=cfg.num_directions,
        layers_modified=stats.get("layers_modified", len(effective_vectors)),
        directions_per_layer=directions_per_layer,
        combination_weights=combination_weights,
        effective_vectors=effective_vectors,
        metadata={"bake_stats": stats})


def train_and_bake_grom(
    model: "Module",
    pair_set: "ContrastivePairSet",
    config: Optional[MultiDirectionConfig] = None,
    grom_config: Optional[Dict[str, Any]] = None,
) -> MultiDirectionResult:
    """Train GROM and bake directions into model weights."""
    from wisent.core.control.steering_methods.methods.grom import GROMMethod, GROMConfig

    if grom_config is None:
        raise ValueError("grom_config dict is required with all GROM hyperparameters")

    cfg = config or MultiDirectionConfig(method="grom")

    # Build GROMConfig from caller-provided dict, overriding num_directions/optimization_steps
    grom_kwargs = dict(grom_config)
    grom_kwargs["num_directions"] = cfg.num_directions
    grom_kwargs["optimization_steps"] = cfg.optimization_steps
    t_cfg = GROMConfig(**grom_kwargs)

    print(f"\nTraining GROM with {cfg.num_directions} directions...")
    grom = GROMMethod(config=t_cfg, log_interval=grom_config["log_interval"])
    result = grom.train_grom(pair_set)

    directions = result.directions
    weights = result.direction_weights

    print(f"Trained {len(directions)} layers")
    for layer, dirs in directions.items():
        print(f"  {layer}: {dirs.shape[0]} directions")

    print(f"\nBaking into weights with strategy: {cfg.combination_strategy}")
    return bake_multi_directions(model, directions, weights, cfg)


# Re-export from helpers
from wisent.core.weight_modification.multi._multi_direction_helpers import (
    train_and_bake_tecza,
    train_and_bake_tetno,
    train_and_bake,
)



"""
LM-Evaluation-Harness task wrapper for task-agnostic architecture.
"""

from typing import Any, Dict, List, Optional

from wisent.core.utils.services.benchmarks import BenchmarkExtractor
from wisent.core.utils.services.benchmarks import get_extractor as get_benchmark_extractor
from wisent.extractors.lm_eval.lm_extractor_registry import get_extractor as get_lm_extractor
from wisent.core.control.tasks.base.task_interface import TaskInterface
from wisent.core.utils import get_test_docs, get_all_docs_from_task, create_deterministic_split


def get_extractor(task_name: str) -> BenchmarkExtractor:
    """Get extractor, trying lm_extractor_registry first, then benchmark_extractors."""
    try:
        return get_lm_extractor(task_name)
    except Exception:
        return get_benchmark_extractor(task_name)


class LMEvalTask(TaskInterface):
    """Wrapper for lm-evaluation-harness tasks."""

    def __init__(self, task_name: str, description: str, categories: List[str], *, train_ratio: float):
        self.task_name = task_name
        self._description = description
        self._categories = categories
        self._extractor = get_extractor(task_name)
        self.train_ratio = train_ratio

    def load_data(self, limit: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Load TEST data from lm-eval using our unified split strategy.

        All available splits are combined and split 80/20 into train/test.
        This method returns the TEST portion for evaluation.
        Training portion is used for contrastive pair generation (see atoms.py).
        """
        try:
            from lm_eval.tasks import get_task_dict

            # Get task directly from lm-eval
            task_dict = get_task_dict([self.task_name])
            if self.task_name not in task_dict:
                print(f"Warning: Task '{self.task_name}' not found in lm-eval")
                return []

            task = task_dict[self.task_name]

            # Get ALL docs from all splits
            all_docs, split_counts = get_all_docs_from_task(task)

            if not all_docs:
                print(f"Warning: No documents found for task '{self.task_name}'")
                return []

            # Apply our 80/20 split and get TEST docs only
            _, test_docs = create_deterministic_split(all_docs, self.task_name, train_ratio=self.train_ratio)

            # Ensure docs are in dictionary format
            processed_docs = []
            for doc in test_docs:
                if isinstance(doc, dict):
                    processed_docs.append(doc)
                elif isinstance(doc, str):
                    processed_docs.append({"text": doc})
                else:
                    try:
                        processed_docs.append(dict(doc))
                    except:
                        processed_docs.append({"data": str(doc)})

            docs = processed_docs

            # Apply limit if specified
            if limit and len(docs) > limit:
                docs = docs[:limit]

            total = len(all_docs)
            test_count = len(test_docs)
            returned = len(docs)
            print(f"Loaded {returned} test docs from {self.task_name} "
                  f"(total: {total}, test split: {test_count}, original splits: {split_counts})")

            return docs

        except Exception as e:
            print(f"Warning: Could not load lm-eval task '{self.task_name}': {e}")
            return []

    def get_extractor(self) -> BenchmarkExtractor:
        """Get the benchmark extractor for this task."""
        return self._extractor

    def get_name(self) -> str:
        """Get the task name."""
        return self.task_name

    def get_description(self) -> str:
        """Get the task description."""
        return self._description

    def get_categories(self) -> List[str]:
        """Get the task categories."""
        return self._categories


class MBPPTask(LMEvalTask):
    """MBPP task implementation."""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="mbpp",
            description="MBPP: Mostly Basic Python Problems coding benchmark",
            categories=["coding", "reasoning", "python"],
            train_ratio=train_ratio,
        )


class HumanEvalTask(LMEvalTask):
    """HumanEval task implementation."""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="humaneval",
            description="HumanEval: Human Evaluation of Python coding problems",
            categories=["coding", "reasoning", "python"],
            train_ratio=train_ratio,
        )


class MBPPPlusTask(LMEvalTask):
    """MBPP Plus task implementation."""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="mbpp_plus",
            description="MBPP Plus: Extended version of MBPP with additional test cases",
            categories=["coding", "reasoning", "python"],
            train_ratio=train_ratio,
        )


class GSM8KTask(LMEvalTask):
    """GSM8K task implementation."""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="gsm8k",
            description="GSM8K: Grade School Math 8K problems",
            categories=["mathematics", "reasoning", "arithmetic"],
            train_ratio=train_ratio,
        )

class ArithmeticBaseTask(LMEvalTask):
    """Base class for arithmetic tasks. Uses unified split strategy."""
    pass  # No longer needs special handling - unified split combines all available docs


class Arithmetic1dcTask(ArithmeticBaseTask):
    """Arithmetic 1dc task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_1dc",
            description="Arithmetic 1dc: 1 digit addition arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )


class Arithmetic2daTask(ArithmeticBaseTask):
    """Arithmetic 2da task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_2da",
            description="Arithmetic 2da: 2 digit addition arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic2dmTask(ArithmeticBaseTask):
    """Arithmetic 2dm task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_2dm",
            description="Arithmetic 2dm: 2 digit multiplication arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic2dsTask(ArithmeticBaseTask):
    """Arithmetic 2ds task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_2ds",
            description="Arithmetic 2ds: 2 digit subtraction arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic3daTask(ArithmeticBaseTask):
    """Arithmetic 3da task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_3da",
            description="Arithmetic 3da: 3 digit addition arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic3dsTask(ArithmeticBaseTask):
    """Arithmetic 3ds task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_3ds",
            description="Arithmetic 3ds: 3 digit subtraction arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic4daTask(ArithmeticBaseTask):
    """Arithmetic 4da task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_4da",
            description="Arithmetic 4da: 4 digit addition arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic4dsTask(ArithmeticBaseTask):
    """Arithmetic 4ds task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_4ds",
            description="Arithmetic 4ds: 4 digit subtraction arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic5daTask(ArithmeticBaseTask):
    """Arithmetic 5da task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_5da",
            description="Arithmetic 5da: 5 digit addition arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )

class Arithmetic5dsTask(ArithmeticBaseTask):
    """Arithmetic 5ds task implementation"""

    def __init__(self, *, train_ratio: float):
        super().__init__(
            task_name="arithmetic_5ds",
            description="Arithmetic 5ds: 5 digit subtraction arithmetic problems",
            categories=["mathematics", "arithmetic"],
            train_ratio=train_ratio,
        )




"""IO and utility mixin for Wisent."""
from __future__ import annotations
from typing import Any, Dict, List, Optional
import logging
import torch

from wisent.core.utils.infra_tools.errors import MissingParameterError

logger = logging.getLogger(__name__)


class WisentIOMixin:
    """Mixin providing IO and utility methods."""

    def act(
        self,
        state: RobotState,
        steer: Dict[str, float] | None = None,
        config: SteeringConfig | None = None,
    ) -> Any:
        """
        Convenience method for robotics: get action from state.

        Args:
            state: Robot state observation
            steer: Steering configuration
            config: Additional config

        Returns:
            Robot action
        """
        if not isinstance(self.adapter, RoboticsAdapter):
            raise AdapterError("act() is only available for robotics adapter")

        return self.generate(state, steer=steer, config=config)

    def transcribe(
        self,
        audio: AudioContent,
        steer: Dict[str, float] | None = None,
        language: str | None = None,
    ) -> str:
        """
        Convenience method for audio: transcribe with steering.

        Args:
            audio: Audio content
            steer: Steering configuration
            language: Target language

        Returns:
            Transcribed text
        """
        if not isinstance(self.adapter, AudioAdapter):
            raise AdapterError("transcribe() is only available for audio adapter")

        if steer is None or not steer:
            return self.adapter.transcribe(audio, language=language)

        combined_vectors = self._combine_steering_vectors(steer)
        return self.adapter.transcribe(
            audio,
            steering_vectors=combined_vectors,
            language=language,
        )

    # ==================== Introspection ====================

    @property
    def traits(self) -> List[str]:
        """Get list of defined traits."""
        return list(self._traits.keys())

    @property
    def is_trained(self) -> bool:
        """Check if steering vectors have been trained."""
        return self._trained

    def get_trait_info(self, trait: str) -> TraitConfig | None:
        """Get configuration for a trait."""
        return self._traits.get(trait)

    def get_intervention_points(self) -> List[str]:
        """Get available steering intervention points."""
        return [ip.name for ip in self.adapter.get_intervention_points()]

    def get_recommended_layers(self) -> List[str]:
        """Get recommended layers for steering."""
        return [
            ip.name for ip in self.adapter.get_intervention_points()
            if ip.recommended
        ]

    # ==================== Persistence ====================

    def save_vectors(self, path: str) -> None:
        """
        Save trained steering vectors to file.

        Args:
            path: File path to save to
        """
        if not self._trained:
            raise NoTrainedVectorsError()

        data = {}
        for trait, config in self._traits.items():
            if config.steering_vectors is not None:
                data[trait] = {
                    "description": config.description,
                    "default_scale": config.default_scale,
                    "layers": config.layers,
                    "vectors": {
                        k: v.cpu() for k, v in config.steering_vectors.items()
                        if v is not None
                    },
                }

        torch.save(data, path)
        logger.info(f"Saved steering vectors to {path}")

    def load_vectors(self, path: str) -> "Wisent":
        """
        Load trained steering vectors from file.

        Args:
            path: File path to load from

        Returns:
            Self for chaining
        """
        data = torch.load(path, map_location=self.device, weights_only=False)

        for trait, trait_data in data.items():
            if "default_scale" not in trait_data:
                raise ValueError(f"default_scale is required in saved vectors for trait '{trait}'")
            self._traits[trait] = TraitConfig(
                name=trait,
                description=trait_data.get("description", ""),
                default_scale=trait_data["default_scale"],
                layers=trait_data.get("layers"),
                steering_vectors=LayerActivations(trait_data["vectors"]),
            )

        self._trained = True
        logger.info(f"Loaded steering vectors from {path}")
        return self

    def __repr__(self) -> str:
        return (
            f"Wisent(\n"
            f"  adapter={self.adapter!r},\n"
            f"  traits={self.traits},\n"
            f"  trained={self._trained}\n"
            f")"
        )



"""Geometry detection type definitions."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Any, List, Tuple


__all__ = [
    "StructureType",
    "StructureScore",
    "GeometryAnalysisConfig",
    "GeometryAnalysisResult",
]


class StructureType(Enum):
    """Types of geometric structures in activation space."""
    LINEAR = "linear"
    CONE = "cone"
    CLUSTER = "cluster"
    MANIFOLD = "manifold"
    SPARSE = "sparse"
    BIMODAL = "bimodal"
    ORTHOGONAL = "orthogonal"
    UNKNOWN = "unknown"


@dataclass
class StructureScore:
    """Score for a single structure type."""
    structure_type: StructureType
    score: float
    confidence: float
    details: Dict[str, Any] = field(default_factory=dict)


@dataclass
class GeometryAnalysisConfig:
    """Configuration for comprehensive geometry analysis."""

    num_components: int = None
    """Number of components/directions to analyze."""

    optimization_steps: int = None
    """Steps for optimization-based methods."""

    min_clusters: int = None
    """Minimum number of clusters."""

    max_clusters: int = None
    """Maximum number of clusters to try."""

    kmeans_max_iterations: int = None
    """Maximum iterations for k-means clustering."""

    manifold_neighbors: int = None
    """Number of neighbors for manifold analysis."""

    def __post_init__(self):
        """Validate required fields."""
        for _n, _v in [("num_components", self.num_components), ("optimization_steps", self.optimization_steps), ("min_clusters", self.min_clusters), ("max_clusters", self.max_clusters), ("manifold_neighbors", self.manifold_neighbors), ("kmeans_max_iterations", self.kmeans_max_iterations)]:
            if _v is None: raise ValueError(f"{_n} is required in GeometryAnalysisConfig")

    use_universal_thresholds: bool = True
    """Whether to use thresholds tuned for universal subspace theory."""


@dataclass
class GeometryAnalysisResult:
    """Results from comprehensive geometry analysis."""

    best_structure: StructureType
    """The structure type that best fits the data."""

    best_score: float
    """Score of the best-fitting structure."""

    all_scores: Dict[str, StructureScore]
    """Scores for all analyzed structure types."""

    recommendation: str
    """Recommended steering method based on geometry."""

    details: Dict[str, Any] = field(default_factory=dict)
    """Additional analysis details."""

    def get_ranking(self) -> List[Tuple[StructureType, float]]:
        """Get structures ranked by score."""
        return sorted(
            [(s.structure_type, s.score) for s in self.all_scores.values()],
            key=lambda x: x[1],
            reverse=True
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Literal

import optuna

from wisent.core.utils.infra_tools.errors import UnknownTypeError
from wisent.core.utils.config_tools.constants import (
    BASE_OPTIMIZER_NAME,
    HYPEROPT_BACKEND_NAME,
    OPTUNA_BACKEND_NAME,
)
from wisent.core.utils.services.optimization.core.parameters import (
    OptimizationRun,
    Param,
)

__all__ = [
    "Direction",
    "HPOConfig",
    "HPORun",
    "BaseOptimizer",
]

Direction = Literal["maximize", "minimize"]


@dataclass(slots=True, frozen=True)
class HPOConfig:
    """
    Configuration for hyperparameter optimization (HPO).

    attributes:
        n_trials:
            number of trials to run.
        direction:
            global default direction ("maximize" or "minimize").
        backend:
            optimizer backend ("optuna" or "hyperopt").
        sampler:
            one of {"tpe", "random", "cmaes"} or None for Optuna default.
        pruner:
            one of {"nop", "median", "sha", "asha", "hyperband"} or None for default.
        study_name:
            optional persistent study name.
        storage:
            Optuna storage URL (e.g., sqlite:///file.db) for persistence.
        seed:
            sampler seed for reproducibility.
        load_if_exists:
            reuse persisted study if it already exists (when storage+study_name set).
    """
    n_trials: int = None
    direction: Direction = "maximize"
    backend: str = OPTUNA_BACKEND_NAME
    sampler: str | None = "tpe"
    pruner: str | None = "asha"
    storage: str | None = None
    study_name: str | None = None
    seed: int | None = None
    load_if_exists: bool = True


@dataclass(slots=True, frozen=True)
class HPORun:
    """Result of an HPO run."""
    study: optuna.Study
    best_params: dict[str, Any]
    best_value: float


class BaseOptimizer:
    """
    Base class for task-agnostic optimizers.

    Supports two usage patterns:
    1. OOP: subclass and implement '_objective(trial)' with trial.suggest_*()
    2. Functional: call 'optimize_fn()' with an objective_fn + Param space dict
    """

    name: str = BASE_OPTIMIZER_NAME
    direction: Direction = "maximize"

    def optimize(self, cfg: HPOConfig) -> HPORun:
        """Run optimization using the OOP subclass pattern."""
        if cfg.n_trials is None:
            raise ValueError("n_trials is required in HPOConfig")
        sampler = self._make_sampler(cfg)
        pruner = self._make_pruner(cfg)
        direction: Direction = getattr(self, "direction", cfg.direction)

        study = optuna.create_study(
            direction=direction,
            sampler=sampler,
            pruner=pruner,
            storage=cfg.storage,
            study_name=cfg.study_name,
            load_if_exists=bool(cfg.storage and cfg.study_name and cfg.load_if_exists),
        )

        study.optimize(self._objective, n_trials=cfg.n_trials, show_progress_bar=False)
        return HPORun(study=study, best_params=study.best_params, best_value=study.best_value)

    def optimize_fn(
        self,
        objective_fn: Callable[[dict[str, Any]], float],
        space: dict[str, Param],
        n_trials: int,
        cfg: HPOConfig | None = None,
    ) -> OptimizationRun:
        """Run optimization using a functional objective + Param space.

        arguments:
            objective_fn: callable that takes a dict of param values and returns a score.
            space: dict mapping param names to Param objects.
            n_trials: number of trials to run.
            cfg: optional HPOConfig for backend/sampler/pruner/storage settings.
        """
        if cfg is None:
            cfg = HPOConfig(n_trials=n_trials, direction=self.direction)
        direction = getattr(self, "direction", cfg.direction)

        from wisent.core.utils.services.optimization.core._backend_converters import (
            run_hyperopt,
            run_optuna_functional,
        )

        if cfg.backend == HYPEROPT_BACKEND_NAME:
            return run_hyperopt(objective_fn, space, n_trials, direction, cfg.seed)

        if cfg.backend == OPTUNA_BACKEND_NAME:
            sampler = self._make_sampler(cfg)
            pruner = self._make_pruner(cfg)
            return run_optuna_functional(
                objective_fn, space, n_trials, direction,
                sampler, pruner, cfg.storage, cfg.study_name, cfg.load_if_exists,
            )

        raise ValueError(f"Unknown backend: {cfg.backend}")

    def _objective(self, trial: optuna.Trial) -> float:
        """Implement one trial; return objective value. Override in subclasses."""
        raise NotImplementedError

    def _make_sampler(self, cfg: HPOConfig) -> optuna.samplers.BaseSampler | None:
        """Create an Optuna sampler based on the config."""
        if cfg.sampler is None:
            return None
        s = cfg.sampler.lower()
        if s == "tpe":
            return optuna.samplers.TPESampler(seed=cfg.seed)
        if s == "random":
            return optuna.samplers.RandomSampler(seed=cfg.seed)
        if s == "cmaes":
            return optuna.samplers.CmaEsSampler(seed=cfg.seed)
        raise UnknownTypeError(entity_type="sampler", value=cfg.sampler, valid_values=["tpe", "random", "cmaes"])

    def _make_pruner(self, cfg: HPOConfig) -> optuna.pruners.BasePruner | None:
        """Create an Optuna pruner based on the config."""
        if cfg.pruner is None:
            return None
        p = cfg.pruner.lower()
        if p == "nop":
            return optuna.pruners.NopPruner()
        if p in {"sha", "asha"}:
            return optuna.pruners.SuccessiveHalvingPruner()
        if p == "median":
            return optuna.pruners.MedianPruner()
        if p == "hyperband":
            return optuna.pruners.HyperbandPruner()
        raise UnknownTypeError(entity_type="pruner", value=cfg.pruner, valid_values=["nop", "sha", "asha", "median", "hyperband"])

    @staticmethod
    def report_and_maybe_prune(trial: optuna.Trial, value: float, step: int) -> None:
        """Report an intermediate metric and prune if the pruner suggests it."""
        trial.report(float(value), step=step)
        if trial.should_prune():
            raise optuna.exceptions.TrialPruned()



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from wisent.core.primitives.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.primitives.contrastive_pairs.core.io.response import NegativeResponse, PositiveResponse
from wisent.extractors.lm_eval.atoms import LMEvalBenchmarkExtractor
from wisent.core.utils.cli.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from lm_eval.api.task import ConfigurableTask


__all__ = ["MlqaExtractor"]
_LOG = setup_logger(__name__)

task_names = (
    "mlqa",
    # Arabic context
    "mlqa_ar_ar", "mlqa_ar_de", "mlqa_ar_vi", "mlqa_ar_zh", "mlqa_ar_en", "mlqa_ar_es", "mlqa_ar_hi",
    # German context
    "mlqa_de_ar", "mlqa_de_de", "mlqa_de_vi", "mlqa_de_zh", "mlqa_de_en", "mlqa_de_es", "mlqa_de_hi",
    # Vietnamese context
    "mlqa_vi_ar", "mlqa_vi_de", "mlqa_vi_vi", "mlqa_vi_zh", "mlqa_vi_en", "mlqa_vi_es", "mlqa_vi_hi",
    # Chinese context
    "mlqa_zh_ar", "mlqa_zh_de", "mlqa_zh_vi", "mlqa_zh_zh", "mlqa_zh_en", "mlqa_zh_es", "mlqa_zh_hi",
    # English context
    "mlqa_en_ar", "mlqa_en_de", "mlqa_en_vi", "mlqa_en_zh", "mlqa_en_en", "mlqa_en_es", "mlqa_en_hi",
    # Spanish context
    "mlqa_es_ar", "mlqa_es_de", "mlqa_es_vi", "mlqa_es_zh", "mlqa_es_en", "mlqa_es_es", "mlqa_es_hi",
    # Hindi context
    "mlqa_hi_ar", "mlqa_hi_de", "mlqa_hi_vi", "mlqa_hi_zh", "mlqa_hi_en", "mlqa_hi_es", "mlqa_hi_hi"
)

class MlqaExtractor(LMEvalBenchmarkExtractor):
    """Extractor for the Mlqa benchmark."""


    evaluator_name = "exact_match"
    def extract_contrastive_pairs(
        self,
        lm_eval_task_data: ConfigurableTask,
        limit: int | None = None,
        preferred_doc: str | None = None,
        *,
        train_ratio: float,
    ) -> list[ContrastivePair]:
        """
        Build contrastive pairs from Mlqa docs.

        Args:
            lm_eval_task_data: lm-eval task instance for Mlqa.
            limit: Optional maximum number of pairs to produce.
            preferred_doc: Optional preferred document source.

        Returns:
            A list of ContrastivePair objects.
        """
        log = bind(_LOG, task=getattr(lm_eval_task_data, "NAME", "unknown"))

        max_items = self._normalize_limit(limit)
        docs = self.load_docs(lm_eval_task_data, max_items, preferred_doc=preferred_doc, train_ratio=train_ratio)

        pairs: list[ContrastivePair] = []

        log.info("Extracting contrastive pairs", extra={"doc_count": len(docs)})

        for doc in docs:
            pair = self._extract_pair_from_doc(doc)
            if pair is not None:
                pairs.append(pair)
                if max_items is not None and len(pairs) >= max_items:
                    break

        if not pairs:
            task_name = getattr(lm_eval_task_data, "NAME", type(lm_eval_task_data).__name__)
            log.warning("No valid Mlqa pairs extracted", extra={"task": task_name})

        return pairs

    def _extract_pair_from_doc(self, doc: dict[str, Any]) -> ContrastivePair | None:
        """
        Convert a single Mlqa doc into a ContrastivePair, if possible.
        Returns None when required fields are missing or malformed.

        Mlqa format: context + question + answers (QA task like SQuAD)
        """
        log = bind(_LOG, doc_id=doc.get("id", "unknown"))

        try:
            # mlqa format: context + question + answers (list of answer texts)
            if "context" in doc and "question" in doc and "answers" in doc:
                context = str(doc.get("context", "")).strip()
                question = str(doc.get("question", "")).strip()
                answers = doc.get("answers")

                # answers can be a dict with "text" key or a list
                if isinstance(answers, dict):
                    answer_texts = answers.get("text", [])
                elif isinstance(answers, list):
                    answer_texts = answers
                else:
                    answer_texts = []

                if not context or not question or not answer_texts:
                    log.debug("Skipping doc with missing context, question, or answers", extra={"doc": doc})
                    return None

                # Use the first answer as the correct answer
                correct_answer = str(answer_texts[0]).strip() if answer_texts else ""

                if not correct_answer:
                    log.debug("Skipping doc with empty answer", extra={"doc": doc})
                    return None

                # Create synthetic negative - generic non-answer for QA
                incorrect_answer = "I don't know."

                prompt = f"Context: {context}\n\nQuestion: {question}\n\nAnswer:"

                metadata = {"label": "mlqa"}
                return self._build_pair(
                    question=prompt,
                    correct=correct_answer,
                    incorrect=incorrect_answer,
                    metadata=metadata,
                )

            log.debug("Skipping doc without context/question/answers fields", extra={"doc": doc})
            return None

        except Exception as exc:
            log.error("Error extracting pair from doc", exc_info=exc, extra={"doc": doc})
            return None

    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(prompt=question, positive_response=positive_response, negative_response=negative_response, label=metadata.get("label"))



"""Backend conversion functions for Hyperopt and Optuna param spaces."""
from __future__ import annotations

import math
from typing import Any, Callable

from wisent.core.utils.config_tools.constants import (
    HYPEROPT_BACKEND_NAME,
    INDEX_FIRST,
    OPTUNA_BACKEND_NAME,
    OPTUNA_SIGMA_SPREAD_FACTOR,
    QUANTIZATION_STEP_DEFAULT,
)
from wisent.core.utils.services.optimization.core.parameters import (
    CategoricalParam,
    FloatParam,
    IntParam,
    OptimizationRun,
    Param,
)


def run_hyperopt(
    objective_fn: Callable[[dict[str, Any]], float],
    space: dict[str, Param],
    n_trials: int,
    direction: str,
    seed: int | None,
) -> OptimizationRun:
    """Run optimization using the Hyperopt backend."""
    from hyperopt import STATUS_OK, Trials, fmin, tpe

    hp_space = {
        name: _param_to_hyperopt(name, param)
        for name, param in space.items()
    }

    maximize = direction == "maximize"
    trials = Trials()

    def wrapped(params):
        score = objective_fn(params)
        return {
            "loss": -score if maximize else score,
            "status": STATUS_OK,
        }

    rstate_kwargs = {}
    if seed is not None:
        import numpy as _np
        rstate_kwargs["rstate"] = _np.random.default_rng(seed)

    fmin(
        fn=wrapped,
        space=hp_space,
        algo=tpe.suggest,
        max_evals=n_trials,
        trials=trials,
        show_progressbar=True,
        **rstate_kwargs,
    )

    all_trials = []
    best_score = None
    best_params = None
    for t in trials.trials:
        vals = t["misc"]["vals"]
        params = {k: v[INDEX_FIRST] for k, v in vals.items()}
        params = resolve_categorical_indices(params, space)
        loss = t["result"]["loss"]
        score = -loss if maximize else loss
        all_trials.append({"params": params, "score": score})
        if best_score is None or (
            (maximize and score > best_score)
            or (not maximize and score < best_score)
        ):
            best_score = score
            best_params = params

    if best_params is None:
        best_raw = {
            k: v[INDEX_FIRST]
            for k, v in trials.best_trial["misc"]["vals"].items()
        }
        best_params = resolve_categorical_indices(best_raw, space)
        best_loss = trials.best_trial["result"]["loss"]
        best_score = -best_loss if maximize else best_loss

    return OptimizationRun(
        best_params=best_params,
        best_score=best_score,
        all_trials=all_trials,
        n_trials=n_trials,
        backend=HYPEROPT_BACKEND_NAME,
    )


def run_optuna_functional(
    objective_fn: Callable[[dict[str, Any]], float],
    space: dict[str, Param],
    n_trials: int,
    direction: str,
    sampler,
    pruner,
    storage: str | None,
    study_name: str | None,
    load_if_exists: bool,
) -> OptimizationRun:
    """Run optimization using Optuna with Param-based space.

    Accepts pre-built sampler and pruner objects from the caller.
    """
    import optuna

    def objective(trial: optuna.Trial) -> float:
        params = {
            name: param_to_optuna(trial, name, param)
            for name, param in space.items()
        }
        return objective_fn(params)

    study = optuna.create_study(
        direction=direction,
        sampler=sampler,
        pruner=pruner,
        storage=storage,
        study_name=study_name,
        load_if_exists=bool(
            storage and study_name and load_if_exists,
        ),
    )
    study.optimize(objective, n_trials=n_trials, show_progress_bar=True)

    all_trials = [
        {"params": t.params, "score": t.value}
        for t in study.trials
        if t.value is not None
    ]

    return OptimizationRun(
        best_params=study.best_params,
        best_score=study.best_value,
        all_trials=all_trials,
        n_trials=n_trials,
        backend=OPTUNA_BACKEND_NAME,
    )


def _param_to_hyperopt(name: str, param: Param):
    """Map a Param dataclass to a hyperopt search expression."""
    from hyperopt import hp

    if isinstance(param, FloatParam):
        return _float_to_hyperopt(name, param)
    if isinstance(param, IntParam):
        return _int_to_hyperopt(name, param)
    if isinstance(param, CategoricalParam):
        return hp.choice(name, param.choices)
    raise TypeError(f"Unknown param type: {type(param)}")


def _float_to_hyperopt(name: str, p: FloatParam):
    from hyperopt import hp

    d = p.distribution
    if d == "normal":
        return hp.normal(name, p.mu, p.sigma)
    if d == "lognormal":
        return hp.lognormal(name, p.mu, p.sigma)
    if d == "uniform":
        return hp.uniform(name, p.low, p.high)
    raise ValueError(f"Unknown float distribution: {d}")


def _int_to_hyperopt(name: str, p: IntParam):
    from hyperopt import hp
    from hyperopt.pyll import scope

    d = p.distribution
    if d == "randint":
        return scope.int(
            hp.quniform(name, p.low, p.high, QUANTIZATION_STEP_DEFAULT),
        )
    if d == "qnormal":
        return scope.int(hp.qnormal(name, p.mu, p.sigma, p.q))
    if d == "qlognormal":
        return scope.int(hp.qlognormal(name, p.mu, p.sigma, p.q))
    raise ValueError(f"Unknown int distribution: {d}")


def param_to_optuna(trial, name: str, param: Param):
    """Map a Param dataclass to an Optuna trial suggestion."""
    if isinstance(param, FloatParam):
        return _float_to_optuna(trial, name, param)
    if isinstance(param, IntParam):
        return _int_to_optuna(trial, name, param)
    if isinstance(param, CategoricalParam):
        return trial.suggest_categorical(name, param.choices)
    raise TypeError(f"Unknown param type: {type(param)}")


def _float_to_optuna(trial, name: str, p: FloatParam):
    d = p.distribution
    if d == "uniform":
        return trial.suggest_float(name, p.low, p.high, log=p.log_scale)
    spread = OPTUNA_SIGMA_SPREAD_FACTOR
    if d == "normal":
        lo = p.mu - spread * p.sigma
        hi = p.mu + spread * p.sigma
        return trial.suggest_float(name, lo, hi)
    if d == "lognormal":
        lo = math.exp(p.mu - spread * p.sigma)
        hi = math.exp(p.mu + spread * p.sigma)
        return trial.suggest_float(name, lo, hi, log=True)
    raise ValueError(f"Unknown float distribution: {d}")


def _int_to_optuna(trial, name: str, p: IntParam):
    d = p.distribution
    if d == "randint":
        return trial.suggest_int(name, p.low, p.high)
    spread = OPTUNA_SIGMA_SPREAD_FACTOR
    if d == "qnormal":
        lo = max(INDEX_FIRST, int(p.mu - spread * p.sigma))
        hi = int(p.mu + spread * p.sigma)
        return trial.suggest_int(name, lo, hi, step=p.q)
    if d == "qlognormal":
        lo = max(
            QUANTIZATION_STEP_DEFAULT,
            int(math.exp(p.mu - spread * p.sigma)),
        )
        hi = int(math.exp(p.mu + spread * p.sigma))
        return trial.suggest_int(name, lo, hi, step=p.q)
    raise ValueError(f"Unknown int distribution: {d}")


def resolve_categorical_indices(
    params: dict, space: dict[str, Param],
) -> dict:
    """Hyperopt returns indices for hp.choice; resolve to values."""
    resolved = {}
    for k, v in params.items():
        if k in space and isinstance(space[k], CategoricalParam):
            idx = int(v)
            choices = space[k].choices
            resolved[k] = choices[idx] if idx < len(choices) else v
        else:
            resolved[k] = v
    return resolved



"""Data loading and activation extraction for mixed concept detection."""

import random
from typing import Dict, List, Tuple

import numpy as np
import torch
from datasets import load_dataset

from wisent.core.primitives.models.wisent_model import WisentModel
from wisent.core.utils.config_tools.constants import DEFAULT_RANDOM_SEED, PROGRESS_LOG_INTERVAL_20


def load_truthfulqa_pairs(n_pairs: int, seed: int = DEFAULT_RANDOM_SEED) -> List[Dict]:
    """Load contrastive pairs from TruthfulQA."""
    random.seed(seed)
    ds = load_dataset("truthfulqa/truthful_qa", "generation", split="validation")
    
    pairs = []
    indices = list(range(len(ds)))
    random.shuffle(indices)
    
    for idx in indices[:n_pairs]:
        sample = ds[idx]
        if sample["incorrect_answers"]:
            pairs.append({
                "question": sample["question"],
                "positive": sample["best_answer"],
                "negative": random.choice(sample["incorrect_answers"]),
                "source": "truthfulqa",
            })
    
    return pairs


def load_hellaswag_pairs(n_pairs: int, seed: int = DEFAULT_RANDOM_SEED) -> List[Dict]:
    """Load contrastive pairs from HellaSwag."""
    random.seed(seed)
    ds = load_dataset("Rowan/hellaswag", split="validation")
    
    pairs = []
    indices = list(range(len(ds)))
    random.shuffle(indices)
    
    for idx in indices[:n_pairs]:
        sample = ds[idx]
        correct_idx = int(sample["label"])
        endings = sample["endings"]
        
        # Get incorrect endings
        incorrect_indices = [i for i in range(len(endings)) if i != correct_idx]
        if incorrect_indices:
            incorrect_idx = random.choice(incorrect_indices)
            
            context = sample["ctx"]
            pairs.append({
                "question": context,
                "positive": endings[correct_idx],
                "negative": endings[incorrect_idx],
                "source": "hellaswag",
            })
    
    return pairs[:n_pairs]


def get_activations(model: WisentModel, text: str, layer: int) -> torch.Tensor:
    """Extract last token activation from a specific layer."""
    inputs = model.tokenizer(text, return_tensors="pt", truncation=True, max_length=model.tokenizer.model_max_length)
    inputs = {k: v.to(model.device) for k, v in inputs.items()}
    
    activations = {}
    
    def hook_fn(module, input, output):
        if isinstance(output, tuple):
            activations["hidden"] = output[0][:, -1, :].detach().cpu()
        else:
            activations["hidden"] = output[:, -1, :].detach().cpu()
    
    layers = model._layers
    handle = layers[layer].register_forward_hook(hook_fn)
    
    with torch.no_grad():
        model.hf_model(**inputs)
    
    handle.remove()
    return activations["hidden"].squeeze(0)


def extract_difference_vectors(
    model: WisentModel,
    pairs: List[Dict],
    layer: int,
    show_progress: bool = True
) -> Tuple[np.ndarray, List[str]]:
    """
    Extract difference vectors (positive - negative) for all pairs.
    
    Returns:
        diff_vectors: [N, hidden_dim] array of difference vectors
        sources: list of source labels (for validation only, not used in detection)
    """
    diff_vectors = []
    sources = []
    
    total = len(pairs)
    for i, pair in enumerate(pairs):
        if show_progress and (i + 1) % PROGRESS_LOG_INTERVAL_20 == 0:
            print(f"  Extracting activations: {i+1}/{total}")
        
        # Format as chat
        prompt = pair["question"]
        
        pos_text = model.tokenizer.apply_chat_template(
            [{"role": "user", "content": prompt}, {"role": "assistant", "content": pair["positive"]}],
            tokenize=False, add_generation_prompt=False
        )
        neg_text = model.tokenizer.apply_chat_template(
            [{"role": "user", "content": prompt}, {"role": "assistant", "content": pair["negative"]}],
            tokenize=False, add_generation_prompt=False
        )
        
        pos_act = get_activations(model, pos_text, layer)
        neg_act = get_activations(model, neg_text, layer)
        
        diff = (pos_act - neg_act).numpy()
        diff_vectors.append(diff)
        sources.append(pair.get("source", "unknown"))
    
    return np.array(diff_vectors), sources



"""Vector generation parser arguments."""

from .generate_vector_parser import setup_generate_vector_parser
from .generate_vector_from_task_parser import setup_generate_vector_from_task_parser
from .generate_vector_from_synthetic_parser import setup_generate_vector_from_synthetic_parser

__all__ = [
    'setup_generate_vector_parser',
    'setup_generate_vector_from_task_parser',
    'setup_generate_vector_from_synthetic_parser',
]
